# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import division
import numpy as np
import pytest
import operator

import odl
from odl.core.set.sets import ComplexNumbers, RealNumbers
from odl.core.util.testutils import (
    all_equal, all_almost_equal, noise_elements, noise_element, simple_fixture)
from odl.core.array_API_support.utils import get_array_and_backend

exponent = simple_fixture('exponent', [2.0, 1.0, float('inf'), 0.5, 1.5])

space_params = ['product_space', 'power_space']
space_ids = [' space={} '.format(p) for p in space_params]

elem_params = ['space', 'real_space', 'numpy_array', 'array', 'scalar',
               '1d_array']
elem_ids = [' element={} '.format(p) for p in elem_params]


@pytest.fixture(scope="module", ids=space_ids, params=space_params)
def space(request, odl_impl_device_pairs):
    name = request.param.strip()
    impl, device = odl_impl_device_pairs
    if name == 'product_space':
        space = odl.ProductSpace(odl.cn(3, impl=impl, device=device),
                                 odl.cn(2, impl=impl, device=device))
    elif name == 'power_space':
        space = odl.ProductSpace(odl.cn(3, impl=impl, device=device), 2)
    else:
        raise ValueError('undefined space')

    return space


@pytest.fixture(scope="module", ids=elem_ids, params=elem_params)
def newpart(request, space):
    element_form = request.param.strip()

    if element_form == 'space':
        tmp = noise_element(space)
        newreal = space.element(tmp.real)
    elif element_form == 'real_space':
        newreal = noise_element(space).real
    elif element_form == 'numpy_array':
        tmp = noise_element(space)
        newreal = [tmp[0].real.asarray(), tmp[1].real.asarray()]
    elif element_form == 'array':
        if space.is_power_space:
            newreal = [[0, 1, 2], [3, 4, 5]]
        else:
            newreal = [[0, 1, 2], [3, 4]]
    elif element_form == 'scalar':
        newreal = np.random.randn()
    elif element_form == '1d_array':
        if not space.is_power_space:
            pytest.skip('arrays matching only one dimension can only be used '
                        'for power spaces')
        newreal = [0, 1, 2]
    else:
        raise ValueError('undefined form of element')

    return newreal


def test_emptyproduct():
    with pytest.raises(ValueError):
        odl.ProductSpace()

    reals = odl.RealNumbers()
    spc = odl.ProductSpace(field=reals)
    assert spc.field == reals
    assert spc.size == 0
    assert spc.is_real
    assert spc.is_complex

    with pytest.raises(IndexError):
        spc[0]


def test_RxR(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    H = odl.rn(2, impl=impl, device=device)
    HxH = odl.ProductSpace(H, H)

    # Check the basic properties
    assert len(HxH) == 2
    assert HxH.shape == (2, 2)
    assert HxH.size == 4
    assert HxH.dtype == H.dtype
    assert HxH.spaces[0] is H
    assert HxH.spaces[1] is H
    assert HxH.is_power_space
    assert not HxH.is_weighted
    assert HxH.is_real
    assert not HxH.is_complex

    v1 = H.element([1, 2])
    v2 = H.element([3, 4])
    v = HxH.element([v1, v2])
    u = HxH.element([[1, 2], [3, 4]])

    assert all_equal([v1, v2], v)
    assert all_equal([v1, v2], u)


def test_equals_space(odl_impl_device_pairs, exponent):
    impl, device = odl_impl_device_pairs
    r2 = odl.rn(2, impl=impl, device=device)
    r2x3_1 = odl.ProductSpace(r2, 3, exponent=exponent)
    r2x3_2 = odl.ProductSpace(r2, 3, exponent=exponent)
    r2x4 = odl.ProductSpace(r2, 4, exponent=exponent)

    assert r2x3_1 is r2x3_1
    assert r2x3_1 is not r2x3_2
    assert r2x3_1 is not r2x4
    assert r2x3_1 == r2x3_1
    assert r2x3_1 == r2x3_2
    assert r2x3_1 != r2x4
    assert hash(r2x3_1) == hash(r2x3_2)
    assert hash(r2x3_1) != hash(r2x4)


def test_equals_vec(odl_impl_device_pairs, exponent):
    impl, device = odl_impl_device_pairs
    r2 = odl.rn(2, impl=impl, device=device)
    r2x3 = odl.ProductSpace(r2, 3, exponent=exponent)
    r2x4 = odl.ProductSpace(r2, 4, exponent=exponent)

    x1 = r2x3.zero()
    x2 = r2x3.zero()
    y = r2x3.one()
    z = r2x4.zero()

    assert x1 is x1
    assert x1 is not x2
    assert x1 is not y
    assert x1 == x1
    assert x1 == x2
    assert x1 != y
    assert x1 != z


def test_is_power_space(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r2 = odl.rn(2, impl=impl, device=device)
    r2x3 = odl.ProductSpace(r2, 3)
    assert len(r2x3) == 3
    assert r2x3.is_power_space
    assert r2x3.spaces[0] is r2
    assert r2x3.spaces[1] is r2
    assert r2x3.spaces[2] is r2

    r2r2r2 = odl.ProductSpace(r2, r2, r2)
    assert r2x3 == r2r2r2


def test_mixed_space(odl_impl_device_pairs):
    """Verify that a mixed productspace is handled properly."""
    impl, device = odl_impl_device_pairs
    r2_1 = odl.rn(2, dtype='float64', impl=impl, device=device)
    r2_2 = odl.rn(2, dtype='float32', impl=impl, device=device)
    pspace = odl.ProductSpace(r2_1, r2_2)

    assert not pspace.is_power_space
    assert pspace.spaces[0] is r2_1
    assert pspace.spaces[1] is r2_2
    assert pspace.is_real
    assert not pspace.is_complex

    # dtype not well defined for this space
    with pytest.raises(AttributeError):
        pspace.dtype


def test_element(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    H = odl.rn(2, impl=impl, device=device)
    HxH = odl.ProductSpace(H, H)

    HxH.element([[1, 2], [3, 4]])

    # wrong length
    with pytest.raises(ValueError):
        # The user tries to input a list of length 1. This would be broadcasted to all parts of the space if cast is True. Hence we need to explicitely set it to False if the strict semantics are desired.
        HxH.element([[1, 2]], cast = False)

    with pytest.raises(ValueError):
        HxH.element([[1, 2], [3, 4], [5, 6]])

    # wrong length of subspace element
    with pytest.raises(ValueError):
        HxH.element([[1, 2, 3], [4, 5]])

    with pytest.raises(ValueError):
        HxH.element([[1, 2], [3, 4, 5]])


def test_lincomb(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    H = odl.rn(2, impl=impl, device=device)
    HxH = odl.ProductSpace(H, H)

    v1 = H.element([1, 2])
    v2 = H.element([5, 3])
    u1 = H.element([-1, 7])
    u2 = H.element([2, 1])

    v = HxH.element([v1, v2])
    u = HxH.element([u1, u2])
    z = HxH.element()

    a = 3.12
    b = 1.23

    expected = [a * v1 + b * u1, a * v2 + b * u2]
    HxH.lincomb(a, v, b, u, out=z)

    assert all_almost_equal(z, expected)


def test_multiply(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    H = odl.rn(2, impl=impl, device=device)
    HxH = odl.ProductSpace(H, H)

    v1 = H.element([1, 2])
    v2 = H.element([5, 3])
    u1 = H.element([-1, 7])
    u2 = H.element([2, 1])

    v = HxH.element([v1, v2])
    u = HxH.element([u1, u2])
    z = HxH.element()

    expected = [v1 * u1, v2 * u2]
    z = v * u

    assert all_almost_equal(z, expected)

    odl.multiply(v, u, out=z)
    assert all_almost_equal(z, expected)



def test_metric(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    H = odl.rn(2, impl=impl, device=device)
    v11 = H.element([1, 2])
    v12 = H.element([5, 3])

    v21 = H.element([1, 2])
    v22 = H.element([8, 9])

    # 1-norm
    HxH = odl.ProductSpace(H, H, exponent=1.0)
    w1 = HxH.element([v11, v12])
    w2 = HxH.element([v21, v22])
    assert (HxH.dist(w1, w2) ==
            pytest.approx(H.dist(v11, v21) + H.dist(v12, v22)))

    # 2-norm
    HxH = odl.ProductSpace(H, H, exponent=2.0)
    w1 = HxH.element([v11, v12])
    w2 = HxH.element([v21, v22])
    assert (
        HxH.dist(w1, w2) ==
        pytest.approx((H.dist(v11, v21) ** 2 + H.dist(v12, v22) ** 2) ** 0.5)
    )

    # inf norm
    HxH = odl.ProductSpace(H, H, exponent=float('inf'))
    w1 = HxH.element([v11, v12])
    w2 = HxH.element([v21, v22])
    assert (HxH.dist(w1, w2) ==
            pytest.approx(max(H.dist(v11, v21), H.dist(v12, v22))))


def test_norm(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    H = odl.rn(2, impl=impl, device=device)
    H = odl.rn(2)
    v1 = H.element([1, 2])
    v2 = H.element([5, 3])

    # 1-norm
    HxH = odl.ProductSpace(H, H, exponent=1.0)
    w = HxH.element([v1, v2])
    assert HxH.norm(w) == pytest.approx(H.norm(v1) + H.norm(v2))

    # 2-norm
    HxH = odl.ProductSpace(H, H, exponent=2.0)
    w = HxH.element([v1, v2])
    assert (HxH.norm(w) ==
            pytest.approx((H.norm(v1) ** 2 + H.norm(v2) ** 2) ** (1 / 2.0)))

    # inf norm
    HxH = odl.ProductSpace(H, H, exponent=float('inf'))
    w = HxH.element([v1, v2])
    assert HxH.norm(w) == pytest.approx(max(H.norm(v1), H.norm(v2)))


def test_inner(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    H = odl.rn(2, impl=impl, device=device)
    v1 = H.element([1, 2])
    v2 = H.element([5, 3])

    u1 = H.element([2, 3])
    u2 = H.element([6, 4])

    HxH = odl.ProductSpace(H, H)
    v = HxH.element([v1, v2])
    u = HxH.element([u1, u2])
    assert HxH.inner(v, u) == pytest.approx(H.inner(v1, u1) + H.inner(v2, u2))


def test_vector_weighting(exponent, odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r2 = odl.rn(2, impl=impl, device=device)
    r2x = r2.element([1, -1])
    r2y = r2.element([-2, 3])
    # inner = -5, dist = 5, norms = (sqrt(2), sqrt(13))

    r3 = odl.rn(3, impl=impl, device=device)
    r3x = r3.element([3, 4, 4])
    r3y = r3.element([1, -2, 1])
    # inner = -1, dist = 7, norms = (sqrt(41), sqrt(6))

    inners = [-5, -1]
    norms_x = [np.sqrt(2), np.sqrt(41)]
    dists = [5, 7]

    weight = [0.5, 1.5]
    pspace = odl.ProductSpace(r2, r3, weighting=weight, exponent=exponent)
    x = pspace.element((r2x, r3x))
    y = pspace.element((r2y, r3y))

    if exponent == 2.0:
        true_inner = np.sum(np.multiply(inners, weight))
        assert all_almost_equal(x.inner(y), true_inner)

    if exponent == float('inf'):
        true_norm_x = np.linalg.norm(
            np.multiply(norms_x, weight), ord=exponent)
    else:
        true_norm_x = np.linalg.norm(
            np.multiply(norms_x, np.power(weight, 1 / exponent)),
            ord=exponent)

    assert all_almost_equal(x.norm(), true_norm_x)

    if exponent == float('inf'):
        true_dist = np.linalg.norm(
            np.multiply(dists, weight), ord=exponent)
    else:
        true_dist = np.linalg.norm(
            np.multiply(dists, np.power(weight, 1 / exponent)),
            ord=exponent)
    assert all_almost_equal(x.dist(y), true_dist)


def test_const_weighting(exponent, odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r2 = odl.rn(2, impl=impl, device=device)
    r2x = r2.element([1, -1])
    r2y = r2.element([-2, 3])
    # inner = -5, dist = 5, norms = (sqrt(2), sqrt(13))

    r3 = odl.rn(3, impl=impl, device=device)
    r3x = r3.element([3, 4, 4])
    r3y = r3.element([1, -2, 1])
    # inner = -1, dist = 7, norms = (sqrt(41), sqrt(6))

    inners = [-5, -1]
    norms_x = [np.sqrt(2), np.sqrt(41)]
    dists = [5, 7]

    weight = 2.0
    pspace = odl.ProductSpace(r2, r3, weighting=weight, exponent=exponent)
    x = pspace.element((r2x, r3x))
    y = pspace.element((r2y, r3y))

    if exponent == 2.0:
        true_inner = weight * np.sum(inners)
        assert all_almost_equal(x.inner(y), true_inner)

    if exponent == float('inf'):
        true_norm_x = weight * np.linalg.norm(norms_x, ord=exponent)
    else:
        true_norm_x = (weight ** (1 / exponent) *
                       np.linalg.norm(norms_x, ord=exponent))

    assert all_almost_equal(x.norm(), true_norm_x)

    if exponent == float('inf'):
        true_dist = weight * np.linalg.norm(dists, ord=exponent)
    else:
        true_dist = (weight ** (1 / exponent) *
                     np.linalg.norm(dists, ord=exponent))

    assert all_almost_equal(x.dist(y), true_dist)


def custom_inner(x1, x2):
    inners = np.fromiter(
        (x1p.inner(x2p) for x1p, x2p in zip(x1.parts, x2.parts)),
        dtype=x1.space[0].dtype_identifier, count=len(x1))

    return x1.space.field.element(np.sum(inners))


def custom_norm(x):
    norms = np.fromiter(
        (xp.norm() for xp in x.parts),
        dtype=x.space[0].dtype_identifier, count=len(x))

    return float(np.linalg.norm(norms, ord=1))


def custom_dist(x1, x2):
    dists = np.fromiter(
        (x1p.dist(x2p) for x1p, x2p in zip(x1.parts, x2.parts)),
        dtype=x1.space[0].dtype_identifier, count=len(x1))

    return float(np.linalg.norm(dists, ord=1))


def test_custom_funcs(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r2 = odl.rn(2, impl=impl, device=device)
    # Checking the standard 1-norm and standard inner product, just to
    # see that the functions are handled correctly.

    r2 = odl.rn(2, impl=impl, device=device)
    r2x = r2.element([1, -1])
    r2y = r2.element([-2, 3])
    # inner = -5, dist = 5, norms = (sqrt(2), sqrt(13))

    r3 = odl.rn(3, impl=impl, device=device)
    r3x = r3.element([3, 4, 4])
    r3y = r3.element([1, -2, 1])
    # inner = -1, dist = 7, norms = (sqrt(41), sqrt(6))

    pspace_2 = odl.ProductSpace(r2, r3, exponent=2.0)
    x = pspace_2.element((r2x, r3x))
    y = pspace_2.element((r2y, r3y))

    pspace_custom = odl.ProductSpace(r2, r3, inner=custom_inner)
    xc = pspace_custom.element((r2x, r3x))
    yc = pspace_custom.element((r2y, r3y))
    assert x.inner(y) == pytest.approx(xc.inner(yc))

    pspace_1 = odl.ProductSpace(r2, r3, exponent=1.0)
    x = pspace_1.element((r2x, r3x))
    y = pspace_1.element((r2y, r3y))

    pspace_custom = odl.ProductSpace(r2, r3, norm=custom_norm)
    xc = pspace_custom.element((r2x, r3x))
    assert x.norm() == pytest.approx(xc.norm())

    pspace_custom = odl.ProductSpace(r2, r3, dist=custom_dist)
    xc = pspace_custom.element((r2x, r3x))
    yc = pspace_custom.element((r2y, r3y))
    assert x.dist(y) == pytest.approx(xc.dist(yc))

    with pytest.raises(TypeError):
        odl.ProductSpace(r2, r3, a=1)  # extra keyword argument

    with pytest.raises(ValueError):
        odl.ProductSpace(r2, r3, norm=custom_norm, inner=custom_inner)

    with pytest.raises(ValueError):
        odl.ProductSpace(r2, r3, dist=custom_dist, inner=custom_inner)

    with pytest.raises(ValueError):
        odl.ProductSpace(r2, r3, norm=custom_norm, dist=custom_dist)

    with pytest.raises(ValueError):
        odl.ProductSpace(r2, r3, norm=custom_norm, exponent=1.0)

    with pytest.raises(ValueError):
        odl.ProductSpace(r2, r3, norm=custom_norm, weighting=2.0)

    with pytest.raises(ValueError):
        odl.ProductSpace(r2, r3, dist=custom_dist, weighting=2.0)

    with pytest.raises(ValueError):
        odl.ProductSpace(r2, r3, inner=custom_inner, weighting=2.0)


def test_power_RxR(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    H = odl.rn(2, impl=impl, device=device)
    HxH = odl.ProductSpace(H, 2)
    assert len(HxH) == 2

    v1 = H.element([1, 2])
    v2 = H.element([3, 4])
    v = HxH.element([v1, v2])
    u = HxH.element([[1, 2], [3, 4]])

    assert all_equal([v1, v2], v)
    assert all_equal([v1, v2], u)


def _test_shape(space, expected_shape):
    """Helper to validate shape of space."""
    space_el = space.element()

    assert space.shape == expected_shape
    assert space_el.shape == expected_shape
    try:
        arr = space_el.asarray()
    except ValueError:
        pass  # could not convert to array
    else:
        assert arr.shape == expected_shape
    assert space.size == np.prod(expected_shape)
    assert space_el.size == np.prod(expected_shape)
    assert len(space) == expected_shape[0]
    assert len(space_el) == expected_shape[0]


def test_power_shape(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r2 = odl.rn(2, impl=impl, device=device)
    r3 = odl.rn(3, impl=impl, device=device)
    """Check if shape and size are correct for higher-order power spaces."""

    empty = odl.ProductSpace(field=odl.RealNumbers())
    empty2 = odl.ProductSpace(r2, 0)
    assert empty.shape == empty2.shape == ()
    assert empty.size == empty2.size == 0

    r2_3 = odl.ProductSpace(r2, 3)
    _test_shape(r2_3, (3, 2))

    r2xr3 = odl.ProductSpace(r2, r3)
    _test_shape(r2xr3, (2,))

    r2xr3_4 = odl.ProductSpace(r2xr3, 4)
    _test_shape(r2xr3_4, (4, 2))

    r2xr3_5_4 = odl.ProductSpace(r2xr3_4, 5)
    _test_shape(r2xr3_5_4, (5, 4, 2))


def test_power_lincomb(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    H = odl.rn(2, impl=impl, device=device)
    HxH = odl.ProductSpace(H, 2)

    v1 = H.element([1, 2])
    v2 = H.element([5, 3])
    u1 = H.element([-1, 7])
    u2 = H.element([2, 1])

    v = HxH.element([v1, v2])
    u = HxH.element([u1, u2])
    z = HxH.element()

    a = 3.12
    b = 1.23

    expected = [a * v1 + b * u1, a * v2 + b * u2]
    HxH.lincomb(a, v, b, u, out=z)

    assert all_almost_equal(z, expected)


def test_power_in_place_modify(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    H = odl.rn(2, impl=impl, device=device)
    HxH = odl.ProductSpace(H, 2)

    v1 = H.element([1, 2])
    v2 = H.element([5, 3])
    u1 = H.element([-1, 7])
    u2 = H.element([2, 1])
    z1 = H.element()
    z2 = H.element()

    v = HxH.element([v1, v2])
    u = HxH.element([u1, u2])
    z = HxH.element([z1, z2])  # z is simply a wrapper for z1 and z2

    a = 3.12
    b = 1.23

    HxH.lincomb(a, v, b, u, out=z)

    # Assert that z1 and z2 has been modified as well
    assert all_almost_equal(z, [z1, z2])


def test_getitem_single(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r1 = odl.rn(1, impl=impl, device=device)
    r2 = odl.rn(2, impl=impl, device=device)
    H = odl.ProductSpace(r1, r2)

    assert H[-2] is r1
    assert H[-1] is r2
    assert H[0] is r1
    assert H[1] is r2
    with pytest.raises(IndexError):
        H[-3]
        H[2]

    assert H[(1,)] == r2
    with pytest.raises(IndexError):
        H[0, 1]


def test_getitem_slice(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r1 = odl.rn(1, impl=impl, device=device)
    r2 = odl.rn(2, impl=impl, device=device)
    r3 = odl.rn(3, impl=impl, device=device)
    H = odl.ProductSpace(r1, r2, r3)

    assert H[:2] == odl.ProductSpace(r1, r2)
    assert H[:2][0] is r1
    assert H[:2][1] is r2

    assert H[3:] == odl.ProductSpace(field=r1.field)


def test_getitem_fancy(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r1 = odl.rn(1, impl=impl, device=device)
    r2 = odl.rn(2, impl=impl, device=device)
    r3 = odl.rn(3, impl=impl, device=device)
    H = odl.ProductSpace(r1, r2, r3)

    assert H[[0, 2]] == odl.ProductSpace(r1, r3)
    assert H[[0, 2]][0] is r1
    assert H[[0, 2]][1] is r3


def test_element_equals(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    H = odl.ProductSpace(
        odl.rn(1, impl=impl, device=device), 
        odl.rn(2, impl=impl, device=device)
        )
    x = H.element([[0], [1, 2]])

    assert x != 0  # test == not always true
    assert x == x

    x_2 = H.element([[0], [1, 2]])
    assert x == x_2

    x_3 = H.element([[3], [1, 2]])
    assert x != x_3

    x_4 = H.element([[0], [1, 3]])
    assert x != x_4


def test_element_getitem_int(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    """Test indexing of product space elements with one or several integers."""
    pspace = odl.ProductSpace(
        odl.rn(1, impl=impl, device=device), 
        odl.rn(2, impl=impl, device=device)
        )

    # One level of product space
    x0 = pspace[0].element([0])
    x1 = pspace[1].element([1, 2])
    x = pspace.element([x0, x1])

    assert x[0] is x0
    assert x[1] is x1
    assert x[-2] is x0
    assert x[-1] is x1
    with pytest.raises(IndexError):
        x[-3]
        x[2]
    assert x[0, 0] == 0
    assert x[1, 0] == 1

    # Two levels of product spaces
    pspace2 = odl.ProductSpace(pspace, 3)
    z = pspace2.element([x, x, x])
    assert z[0] is x
    assert z[1, 0] is x0
    assert z[1, 1, 1] == 2


def test_element_getitem_slice(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    """Test indexing of product space elements with slices."""
    # One level of product space
    pspace = odl.ProductSpace(
        odl.rn(1, impl=impl, device=device), 
        odl.rn(2, impl=impl, device=device), 
        odl.rn(3, impl=impl, device=device)
        )

    x0 = pspace[0].element([0])
    x1 = pspace[1].element([1, 2])
    x2 = pspace[2].element([3, 4, 5])
    x = pspace.element([x0, x1, x2])

    assert x[:2].space == pspace[:2]
    assert x[:2][0] is x0
    assert x[:2][1] is x1


def test_element_getitem_fancy(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    pspace = odl.ProductSpace(
        odl.rn(1, impl=impl, device=device),  
        odl.rn(2, impl=impl, device=device),  
        odl.rn(3, impl=impl, device=device)
        )

    x0 = pspace[0].element([0])
    x1 = pspace[1].element([1, 2])
    x2 = pspace[2].element([3, 4, 5])
    x = pspace.element([x0, x1, x2])

    assert x[[0, 2]].space == pspace[[0, 2]]
    assert x[[0, 2]][0] is x0
    assert x[[0, 2]][1] is x2


def test_element_getitem_multi(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    """Test element access with multiple indices."""
    pspace = odl.ProductSpace(
        odl.rn(1, impl=impl, device=device),  
        odl.rn(2, impl=impl, device=device)
        )
    pspace2 = odl.ProductSpace(pspace, 3)
    pspace3 = odl.ProductSpace(pspace2, 2)
    z = pspace3.element(
        [[[[1],
           [2, 3]],
          [[4],
           [5, 6]],
          [[7],
           [8, 9]]],
         [[[10],
           [12, 13]],
          [[14],
           [15, 16]],
          [[17],
           [18, 19]]]
         ]
    )

    assert pspace3.shape == (2, 3, 2)
    assert z[0, 0, 0, 0] == 1
    assert all_equal(z[0, 0, 1], [2, 3])
    assert all_equal(z[0, 0], [[1], [2, 3]])
    assert all_equal(z[0, 1:], [[[4],
                                 [5, 6]],
                                [[7],
                                 [8, 9]]])
    assert all_equal(z[0, 1:, 1], [[5, 6],
                                   [8, 9]])
    assert all_equal(z[0, 1:, :, 0], [[[4],
                                       [5]],
                                      [[7],
                                       [8]]])


def test_element_setitem_single(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    """Test assignment of pspace parts with single indices."""
    pspace = odl.ProductSpace(
        odl.rn(1, impl=impl, device=device),   
        odl.rn(2, impl=impl, device=device),  
        )

    x0 = pspace[0].element([0])
    x1 = pspace[1].element([1, 2])
    x = pspace.element([x0, x1])
    old_x0 = x[0]
    old_x1 = x[1]

    # Check that values are set, but identity is preserved
    new_x0 = pspace[0].element([1])
    x[-2] = new_x0
    assert x[-2] == new_x0
    assert x[-2] is old_x0

    new_x1 = pspace[1].element([3, 4])
    x[-1] = new_x1
    assert x[-1] == new_x1
    assert x[-1] is old_x1

    # Set values with scalars
    x[1] = -1
    assert all_equal(x[1], [-1, -1])
    assert x[1] is old_x1

    # Check that out-of-bounds indices raise IndexError
    with pytest.raises(IndexError):
        x[-3] = x1
    with pytest.raises(IndexError):
        x[2] = x0


def test_element_setitem_slice(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    """Test assignment of pspace parts with slices."""
    pspace = odl.ProductSpace(
        odl.rn(1, impl=impl, device=device),   
        odl.rn(2, impl=impl, device=device),   
        odl.rn(3, impl=impl, device=device),  
        )

    x0 = pspace[0].element([0])
    x1 = pspace[1].element([1, 2])
    x2 = pspace[2].element([3, 4, 5])
    x = pspace.element([x0, x1, x2])
    old_x0 = x[0]
    old_x1 = x[1]

    # Check that values are set, but identity is preserved
    new_x0 = pspace[0].element([6])
    new_x1 = pspace[1].element([7, 8])
    x[:2] = pspace[:2].element([new_x0, new_x1])
    assert x[:2][0] is old_x0
    assert x[:2][0] == new_x0
    assert x[:2][1] is old_x1
    assert x[:2][1] == new_x1

    # Set values with sequences of scalars
    x[:2] = [-1, -2]
    assert x[:2][0] is old_x0
    assert all_equal(x[:2][0], [-1])
    assert x[:2][1] is old_x1
    assert all_equal(x[:2][1], [-2, -2])


def test_element_setitem_fancy(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    """Test assignment of pspace parts with lists."""
    pspace = odl.ProductSpace(
        odl.rn(1, impl=impl, device=device),   
        odl.rn(2, impl=impl, device=device),  
        odl.rn(3, impl=impl, device=device),  
        )

    x0 = pspace[0].element([0])
    x1 = pspace[1].element([1, 2])
    x2 = pspace[2].element([3, 4, 5])
    x = pspace.element([x0, x1, x2])
    old_x0 = x[0]
    old_x2 = x[2]

    # Check that values are set, but identity is preserved
    new_x0 = pspace[0].element([6])
    new_x2 = pspace[2].element([7, 8, 9])
    x[[0, 2]] = pspace[[0, 2]].element([new_x0, new_x2])
    assert x[[0, 2]][0] is old_x0
    assert x[[0, 2]][0] == new_x0
    assert x[[0, 2]][1] is old_x2
    assert x[[0, 2]][1] == new_x2

    # Set values with sequences of scalars
    x[[0, 2]] = [-1, -2]
    assert x[[0, 2]][0] is old_x0
    assert all_equal(x[[0, 2]][0], [-1])
    assert x[[0, 2]][1] is old_x2
    assert all_equal(x[[0, 2]][1], [-2, -2, -2])


def test_element_setitem_broadcast(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    """Test assignment of power space parts with broadcasting."""
    pspace = odl.ProductSpace(
        odl.rn(2, impl=impl, device=device),   
        3)
    x0 = pspace[0].element([0, 1])
    x1 = pspace[1].element([2, 3])
    x2 = pspace[2].element([4, 5])
    x = pspace.element([x0, x1, x2])
    old_x0 = x[0]
    old_x1 = x[1]

    # Set values with a single base space element
    new_x0 = pspace[0].element([4, 5])
    x[:2] = new_x0
    assert x[0] is old_x0
    assert x[0] == new_x0
    assert x[1] is old_x1
    assert x[1] == new_x0

def test_unary_ops(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    # Verify that the unary operators (`+x` and `-x`) work as expected
    space = odl.rn(3, impl=impl, device=device)
    pspace = odl.ProductSpace(space, 2)

    for op in [operator.pos, operator.neg]:
        x_arr, x = noise_elements(pspace)

        y_arr = [op(x_) for x_ in x_arr]
        y = op(x)

        assert all_almost_equal([x, y], [x_arr, y_arr])

def test_operators(odl_arithmetic_op, odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    # Test of the operators `+`, `-`, etc work as expected by numpy

    op = odl_arithmetic_op

    space = odl.rn(3, impl=impl, device=device)
    pspace = odl.ProductSpace(space, 2)

    # Interactions with scalars

    for scalar in [-31.2, -1, 0, 1, 2.13]:

        # Left op
        x_arr, x = noise_elements(pspace)
        if scalar == 0 and op in [operator.truediv, operator.itruediv]:
            # Check for correct zero division behaviour
            with pytest.raises(ZeroDivisionError):
                y = op(x, scalar)
        else:
            y_arr = [op(x_, scalar) for x_ in x_arr]
            y = op(x, scalar)

            assert all_almost_equal([x, y], [x_arr, y_arr])

        # Right op
        x_arr, x = noise_elements(pspace)

        y_arr = [op(scalar, x_) for x_ in x_arr]
        y = op(scalar, x)

        assert all_almost_equal([x, y], [x_arr, y_arr])

    # Verify that the statement z=op(x, y) gives equivalent results to NumPy
    x_arr, x = noise_elements(space, 1)
    y_arr, y = noise_elements(pspace, 1)

    # non-aliased left
    if op in [operator.iadd, operator.isub, operator.itruediv, operator.imul]:
        # Check for correct error since in-place op is not possible here
        with pytest.raises(TypeError):
            z = op(x, y)
    else:
        z_arr = [op(x_arr, y_) for y_ in y_arr]
        z = op(x, y)

        assert all_almost_equal([x, y, z], [x_arr, y_arr, z_arr])

    # non-aliased right
    z_arr = [op(y_, x_arr) for y_ in y_arr]
    z = op(y, x)
    assert all_almost_equal([x, y, z], [x_arr, y_arr, z_arr])

    # aliased operation
    z_arr = [op(y_arr[i], y_arr[i]) for i in range(len(y_arr))]
    z = op(y, y)

    assert all_almost_equal([x, y, z], [x_arr, y_arr, z_arr])


def test_ufuncs(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    # Cannot use fixture due to bug in pytest
    H = odl.ProductSpace(
        odl.rn(1, impl=impl, device=device),    
        odl.rn(2, impl=impl, device=device),   
        )

    # one arg
    x = H.element([[-1], [-2, -3]])

    z = odl.abs(x)
    assert all_almost_equal(z, [[1], [2, 3]])

    # one arg with out
    x = H.element([[-1], [-2, -3]])
    y = H.element()

    z = odl.abs(x, out=y)
    assert y is z
    assert all_almost_equal(z, [[1], [2, 3]])

    # Two args
    x = H.element([[1], [2, 3]])
    y = H.element([[4], [5, 6]])
    w = H.element()

    z = odl.add(x, y)
    assert all_almost_equal(z, [[5], [7, 9]])

    # Two args with out
    x = H.element([[1], [2, 3]])
    y = H.element([[4], [5, 6]])
    w = H.element()

    z = odl.add(x, y, out=w)
    assert w is z
    assert all_almost_equal(z, [[5], [7, 9]])


def test_reductions(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    H = odl.ProductSpace(
        odl.rn(1, impl=impl, device=device),    
        odl.rn(2, impl=impl, device=device),    
        )
    x = H.element([[1], [2, 3]])
    assert odl.sum(x) == 6.0
    assert odl.prod(x) == 6.0
    assert odl.min(x) == 1.0
    assert odl.max(x) == 3.0



def test_array_wrap_method(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    """Verify that the __array_wrap__ method for NumPy works."""
    sub_space = odl.rn(10, impl=impl, device=device)
    space = odl.ProductSpace(sub_space, 2)
    x_arr, x = noise_elements(space)

    y_arr = [sub_space.array_namespace.sin(sub_part) for sub_part in x_arr]
    y = odl.sin(x)  # Should yield again an ODL product space element

    assert y in space
    assert all_equal(y, y_arr)


def test_real_imag_and_conj(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    """Verify that .real .imag and .conj() work for product space elements."""
    space = odl.ProductSpace(
        odl.cn(3, impl=impl, device=device),
        odl.cn(2, impl=impl, device=device)
        )
    x = noise_element(space)

    # Test real
    expected_result = space.real_space.element([x[0].real, x[1].real])
    assert x.real == expected_result

    # Test imag
    expected_result = space.real_space.element([x[0].imag, x[1].imag])
    assert x.imag == expected_result

    # Test conj. Note that ProductSpace does not implement asarray if
    # is_power_space is false. Hence the construction below
    expected_result = space.element([x[0].conj(), x[1].conj()])
    x_conj = x.conj()
    assert x_conj[0] == expected_result[0]
    assert x_conj[1] == expected_result[1]


# def test_real_setter_product_space(space, newpart):
#     """Verify that the setter for the real part of an element works.
#     What setting the real part means depends on the inputs; we perform a
#     recursive deconstruction to cover the possible cases.
#     Barring deeply nested products, the recursion will only be shallow
#     (depth 2 for a simple product space). We limit it to a depth of at
#     most 4, to avoid that if some bug causes an infinite recursion,
#     the user would get a cryptic stack-overflow error."""

#     if getattr(newpart, 'space', odl.rn(1)).field == ComplexNumbers():
#         # It is not possible to set a real part to a complex number, skip this case
#         return

#     def verify_result(x, expected_result, recursion_limit=4):
#         if recursion_limit <= 0:
#             return False
#         try:
#             # Catch scalar argument
#             iter(expected_result)
#         except TypeError:
#             return verify_result(x, expected_result * space.one(),
#                                  recursion_limit - 1)
#         if expected_result in space:
#             return all_equal(x.real, expected_result.real)
#         elif all_equal(x.real, expected_result):
#             return True
#         elif space.is_power_space:
#             return verify_result(x, [expected_result for _ in space],
#                                  recursion_limit - 1)

#     x = noise_element(space)
#     x.real = newpart

#     assert x in space
#     assert(verify_result(x, newpart))

#     return


# def test_imag_setter_product_space(space, newpart):
#     """Like test_real_setter_product_space but for imaginary part."""

#     if getattr(newpart, 'space', odl.rn(1)).field == ComplexNumbers():
#         # The imaginary part is itself a real quantity, and
#         # cannot be set to a complex value. Skip test.
#         return

#     def verify_result(x, expected_result, recursion_limit=4):
#         if recursion_limit <= 0:
#             return False
#         try:
#             # Catch scalar argument
#             iter(expected_result)
#         except TypeError:
#             return verify_result(x, expected_result * space.one(),
#                                  recursion_limit - 1)
#         if expected_result in space:
#             # The imaginary part is by definition real, and thus the new
#             # imaginary part is thus the real part of the element we try to set
#             # the value to
#             return all_equal(x.imag, expected_result.real)
#         elif all_equal(x.imag, expected_result):
#             return True
#         elif space.is_power_space:
#             return verify_result(x, [expected_result for _ in space],
#                                  recursion_limit - 1)

#     x = noise_element(space)
#     x.imag = newpart

#     assert x in space
#     assert(verify_result(x, newpart))

#     return


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Unit tests for `tensor_ops`."""

from __future__ import division

import numpy as np
import scipy.sparse

import odl
import pytest
from odl.core.operator.tensor_ops import (
    MatrixOperator, PointwiseInner, PointwiseNorm, PointwiseSum)
from odl.core.space.pspace import ProductSpace
from odl.core.util.testutils import (
    all_almost_equal, all_equal, noise_element, noise_elements, simple_fixture, skip_if_no_pytorch)
from odl.core.space.entry_points import tensor_space_impl_names
from odl.core.sparse import SparseMatrix
from odl.core.array_API_support import lookup_array_backend, get_array_and_backend

matrix_dtype = simple_fixture(
    name='matrix_dtype',
    params=['float32', 'complex64', 'float64', 'complex128'])


@pytest.fixture(scope='module')
def matrix(matrix_dtype, odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    space = odl.rn((3, 4), impl=impl, device=device, dtype=matrix_dtype)
    return space.one()

    # else:
    #     assert 0


exponent = simple_fixture('exponent', [2.0, 1.0, float('inf'), 3.5, 1.5])

sparse_matrix_backend = simple_fixture('backend', ['scipy', 'pytorch'])
sparse_matrix_format = simple_fixture('format', ['COO'])
# ---- PointwiseNorm ----


def test_pointwise_norm_init_properties(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    # 1d
    fspace = odl.uniform_discr([0, 0], [1, 1], (2, 2), impl=impl, device=device)
    vfspace = ProductSpace(fspace, 1, exponent=1)

    # Make sure the code runs and test the properties
    pwnorm = PointwiseNorm(vfspace)
    assert pwnorm.base_space == fspace
    assert all_equal(pwnorm.weights, [1])
    assert not pwnorm.is_weighted
    assert pwnorm.exponent == 1.0
    repr(pwnorm)

    pwnorm = PointwiseNorm(vfspace, exponent=2)
    assert pwnorm.exponent == 2

    pwnorm = PointwiseNorm(vfspace, weighting=2)
    assert all_equal(pwnorm.weights, [2])
    assert pwnorm.is_weighted

    # 3d
    fspace = odl.uniform_discr([0, 0], [1, 1], (2, 2), impl=impl, device=device)
    vfspace = ProductSpace(fspace, 3, exponent=1)

    # Make sure the code runs and test the properties
    pwnorm = PointwiseNorm(vfspace)
    assert pwnorm.base_space == fspace
    assert all_equal(pwnorm.weights, [1, 1, 1])
    assert not pwnorm.is_weighted
    assert pwnorm.exponent == 1.0
    repr(pwnorm)

    pwnorm = PointwiseNorm(vfspace, exponent=2)
    assert pwnorm.exponent == 2

    pwnorm = PointwiseNorm(vfspace, weighting=[1, 2, 3])
    assert all_equal(pwnorm.weights, [1, 2, 3])
    assert pwnorm.is_weighted

    # Bad input
    with pytest.raises(TypeError):
        PointwiseNorm(odl.rn(3))  # No power space

    with pytest.raises(ValueError):
        PointwiseNorm(vfspace, exponent=0.5)  # < 1 not allowed

    with pytest.raises(ValueError):
        PointwiseNorm(vfspace, weighting=-1)  # < 0 not allowed

    with pytest.raises(ValueError):
        PointwiseNorm(vfspace, weighting=[1, 0, 1])  # 0 invalid


def test_pointwise_norm_real(exponent, odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    # 1d
    fspace = odl.uniform_discr([0, 0], [1, 1], (2, 2), impl=impl, device=device)
    vfspace = ProductSpace(fspace, 1)
    pwnorm = PointwiseNorm(vfspace, exponent)

    testarr = fspace.array_backend.array_constructor([[[1, 2],
                         [3, 4]]], dtype=float, device=device)

    true_norm = fspace.array_namespace.linalg.norm(testarr, ord=exponent, axis=0)

    func = vfspace.element(testarr)
    func_pwnorm = pwnorm(func)
    assert all_almost_equal(func_pwnorm, true_norm)

    out = fspace.element()
    pwnorm(func, out=out)
    assert all_almost_equal(out, true_norm)

    # 3d
    fspace = odl.uniform_discr([0, 0], [1, 1], (2, 2), impl=impl, device=device)
    vfspace = ProductSpace(fspace, 3)
    pwnorm = PointwiseNorm(vfspace, exponent)

    testarr = fspace.array_backend.array_constructor([[[1, 2],
                         [3, 4]],
                        [[0, -1],
                         [0, 1]],
                        [[1, 1],
                         [1, 1]]], dtype=float, device=device)

    true_norm = fspace.array_namespace.linalg.norm(testarr, ord=exponent, axis=0)

    func = vfspace.element(testarr)
    func_pwnorm = pwnorm(func)
    assert all_almost_equal(func_pwnorm, true_norm)

    out = fspace.element()
    pwnorm(func, out=out)
    assert all_almost_equal(out, true_norm)


def test_pointwise_norm_complex(exponent, odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    fspace = odl.uniform_discr([0, 0], [1, 1], (2, 2), dtype=complex, impl=impl, device=device)
    vfspace = ProductSpace(fspace, 3)
    pwnorm = PointwiseNorm(vfspace, exponent)

    testarr = fspace.array_backend.array_constructor([[[1 + 1j, 2],
                         [3, 4 - 2j]],
                        [[0, -1],
                         [0, 1]],
                        [[1j, 1j],
                         [1j, 1j]]], device=device, dtype=complex)

    true_norm = fspace.array_namespace.linalg.norm(testarr, ord=exponent, axis=0)

    func = vfspace.element(testarr)
    func_pwnorm = pwnorm(func)
    assert all_almost_equal(func_pwnorm, true_norm)

    out = pwnorm.range.element()
    pwnorm(func, out=out)
    assert all_almost_equal(out.real, true_norm)


def test_pointwise_norm_weighted(exponent, odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    fspace = odl.uniform_discr([0, 0], [1, 1], (2, 2), impl=impl, device=device)

    ns = fspace.array_namespace
    backend = fspace.array_backend

    vfspace = ProductSpace(fspace, 3)
    weight = backend.array_constructor([1.0, 2.0, 3.0], device=device)
    pwnorm = PointwiseNorm(vfspace, exponent, weighting=weight)

    testarr = backend.array_constructor([[[1, 2],
                         [3, 4]],
                        [[0, -1],
                         [0, 1]],
                        [[1, 1],
                         [1, 1]]], device=device, dtype=float)

    if exponent in (1.0, float('inf')):
        true_norm = ns.linalg.norm(weight[:, None, None] * testarr,
                                   ord=exponent, axis=0)
    else:
        true_norm = ns.linalg.norm(
            weight[:, None, None] ** (1 / exponent) * testarr, ord=exponent,
            axis=0)

    func = vfspace.element(testarr)
    func_pwnorm = pwnorm(func)
    assert all_almost_equal(func_pwnorm, true_norm)

    out = fspace.element()
    pwnorm(func, out=out)
    assert all_almost_equal(out, true_norm)


def test_pointwise_norm_gradient_real(exponent, odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    # The operator is not differentiable for exponent 'inf'
    if exponent == float('inf'):
        fspace = odl.uniform_discr([0, 0], [1, 1], (2, 2), impl=impl, device=device)
        vfspace = ProductSpace(fspace, 1)
        pwnorm = PointwiseNorm(vfspace, exponent)
        point = vfspace.one()
        with pytest.raises(NotImplementedError):
            pwnorm.derivative(point)
        return

    # 1d
    fspace = odl.uniform_discr([0, 0], [1, 1], (2, 2), impl=impl, device=device)
    vfspace = ProductSpace(fspace, 1)
    pwnorm = PointwiseNorm(vfspace, exponent)

    point = noise_element(vfspace)
    direction = noise_element(vfspace)

    # Computing expected result
    tmp = odl.pow(pwnorm(point), 1 - exponent)
    v_field = vfspace.element()
    for i in range(len(v_field)):
        v_field[i] = tmp * point[i] * odl.abs(point[i]) ** (exponent - 2)
    pwinner = odl.PointwiseInner(vfspace, v_field)
    expected_result = pwinner(direction)

    func_pwnorm = pwnorm.derivative(point)

    assert all_almost_equal(func_pwnorm(direction), expected_result)

    # 3d
    fspace = odl.uniform_discr([0, 0], [1, 1], (2, 2), impl=impl, device=device)
    vfspace = ProductSpace(fspace, 3)
    pwnorm = PointwiseNorm(vfspace, exponent)

    point = noise_element(vfspace)
    direction = noise_element(vfspace)

    # Computing expected result
    tmp = odl.pow(pwnorm(point), 1 - exponent)
    v_field = vfspace.element()
    for i in range(len(v_field)):
        v_field[i] = tmp * point[i] * odl.abs(point[i]) ** (exponent - 2)
    pwinner = odl.PointwiseInner(vfspace, v_field)
    expected_result = pwinner(direction)

    func_pwnorm = pwnorm.derivative(point)
    assert all_almost_equal(func_pwnorm(direction), expected_result)


def test_pointwise_norm_gradient_real_with_zeros(
        exponent,
        odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    # The gradient is only well-defined in points with zeros if the exponent is
    # >= 2 and < inf
    if exponent < 2 or exponent == float('inf'):
        pytest.skip('differential of operator has singularity for this '
                    'exponent')

    # 1d
    fspace = odl.uniform_discr([0, 0], [1, 1], (2, 2), impl=impl, device=device)
    vfspace = ProductSpace(fspace, 1)
    pwnorm = PointwiseNorm(vfspace, exponent)

    backend = fspace.array_backend

    # This makes the point singular for p < 2
    test_point = backend.array_constructor(
        [[[0, 0],  [1, 2]]], device=device)
    test_direction = backend.array_constructor(
        [[[1, 2], [4, 5]]], device=device)

    point = vfspace.element(test_point)
    direction = vfspace.element(test_direction)
    func_pwnorm = pwnorm.derivative(point)

    assert not odl.any(odl.isnan(func_pwnorm(direction)))

    # 3d
    fspace = odl.uniform_discr([0, 0], [1, 1], (2, 2), impl=impl, device=device)
    vfspace = ProductSpace(fspace, 3)
    pwnorm = PointwiseNorm(vfspace, exponent)

    # This makes the point singular for p < 2
    test_point = backend.array_constructor(
        [[[0, 0],  
        [1, 2]],
        [[3, 4],
        [0, 0]], 
        [[5, 6],
        [7, 8]]], device=device)
    test_direction = backend.array_constructor(
        [[[0, 1],
        [2, 3]],
        [[4, 5],
        [6, 7]],
        [[8, 9],
        [0, 1]]], device=device)

    point = vfspace.element(test_point)
    direction = vfspace.element(test_direction)
    func_pwnorm = pwnorm.derivative(point)

    assert not odl.any(odl.isnan(func_pwnorm(direction)))

# ---- PointwiseInner ----


def test_pointwise_inner_init_properties(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    fspace = odl.uniform_discr([0, 0], [1, 1], (2, 2), impl=impl, device=device)
    vfspace = ProductSpace(fspace, 3, exponent=2)

    # Make sure the code runs and test the properties
    pwinner = PointwiseInner(vfspace, vfspace.one())
    assert pwinner.base_space == fspace
    assert all_equal(pwinner.weights, [1, 1, 1])
    assert not pwinner.is_weighted
    repr(pwinner)

    pwinner = PointwiseInner(vfspace, vfspace.one(), weighting=[1, 2, 3])
    assert all_equal(pwinner.weights, [1, 2, 3])
    assert pwinner.is_weighted

    # Bad input
    with pytest.raises(TypeError):
        PointwiseInner(odl.rn(3), odl.rn(3).one())  # No power space

    # TODO: Does not raise currently, although bad_vecfield not in vfspace!
    """
    bad_vecfield = ProductSpace(fspace, 3, exponent=1).one()
    with pytest.raises(TypeError):
        PointwiseInner(vfspace, bad_vecfield)
    """


def test_pointwise_inner_real(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    # 1d
    fspace = odl.uniform_discr([0, 0], [1, 1], (2, 2), impl=impl, device=device)
    
    backend = fspace.array_backend
    
    vfspace = ProductSpace(fspace, 1)
    array = backend.array_constructor(
        [[[-1, -3], [2, 0]]], device=device)

    pwinner = PointwiseInner(vfspace, vecfield=array)

    testarr = backend.array_constructor(
        [[[1, 2], [3, 4]]], device=device)

    true_inner = backend.array_namespace.sum(testarr * array, axis=0)

    func = vfspace.element(testarr)
    func_pwinner = pwinner(func)
    assert all_almost_equal(func_pwinner, true_inner)

    out = fspace.element()
    pwinner(func, out=out)
    assert all_almost_equal(out, true_inner)

    # 3d
    fspace = odl.uniform_discr([0, 0], [1, 1], (2, 2), impl=impl, device=device)
    vfspace = ProductSpace(fspace, 3)
    array =  backend.array_constructor([[[-1, -3],
                       [2, 0]],
                      [[0, 0],
                       [0, 1]],
                      [[-1, 1],
                       [1, 1]]], device=device)
    pwinner = PointwiseInner(vfspace, vecfield=array)

    testarr =  backend.array_constructor([[[1, 2],
                         [3, 4]],
                        [[0, -1],
                         [0, 1]],
                        [[1, 1],
                         [1, 1]]], device=device)

    true_inner =  backend.array_namespace.sum(testarr * array, axis=0)

    func = vfspace.element(testarr)
    func_pwinner = pwinner(func)
    assert all_almost_equal(func_pwinner, true_inner)

    out = fspace.element()
    pwinner(func, out=out)
    assert all_almost_equal(out, true_inner)


def test_pointwise_inner_complex(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    fspace = odl.uniform_discr([0, 0], [1, 1], (2, 2), dtype=complex, impl=impl, device=device)
    vfspace = ProductSpace(fspace, 3)

    backend = fspace.array_backend

    array = backend.array_constructor([[[-1 - 1j, -3],
                       [2, 2j]],
                      [[-1j, 0],
                       [0, 1]],
                      [[-1, 1 + 2j],
                       [1, 1]]], device=device)
    
    pwinner = PointwiseInner(vfspace, vecfield=array)

    testarr = backend.array_constructor([[[1 + 1j, 2],
                         [3, 4 - 2j]],
                        [[0, -1],
                         [0, 1]],
                        [[1j, 1j],
                         [1j, 1j]]], device=device)

    true_inner = backend.array_namespace.sum(testarr * array.conj(), axis=0)

    func = vfspace.element(testarr)
    func_pwinner = pwinner(func)
    assert all_almost_equal(func_pwinner, true_inner)

    out = fspace.element()
    pwinner(func, out=out)
    assert all_almost_equal(out, true_inner)


def test_pointwise_inner_weighted(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    fspace = odl.uniform_discr([0, 0], [1, 1], (2, 2), impl=impl, device=device)

    backend = fspace.array_backend

    vfspace = ProductSpace(fspace, 3)
    array = backend.array_constructor([[[-1, -3],
                       [2, 0]],
                      [[0, 0],
                       [0, 1]],
                      [[-1, 1],
                       [1, 1]]], device=device)

    weight = backend.array_constructor([1.0, 2.0, 3.0], device=device)
    pwinner = PointwiseInner(vfspace, vecfield=array, weighting=weight)

    testarr = backend.array_constructor([[[1, 2],
                         [3, 4]],
                        [[0, -1],
                         [0, 1]],
                        [[1, 1],
                         [1, 1]]], device=device)

    true_inner = backend.array_namespace.sum(weight[:, None, None] * testarr * array, axis=0)

    func = vfspace.element(testarr)
    func_pwinner = pwinner(func)
    assert all_almost_equal(func_pwinner, true_inner)

    out = fspace.element()
    pwinner(func, out=out)
    assert all_almost_equal(out, true_inner)


def test_pointwise_inner_adjoint(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    # 1d
    fspace = odl.uniform_discr([0, 0], [1, 1], (2, 2), dtype=complex, impl=impl, device=device)

    backend = fspace.array_backend

    vfspace = ProductSpace(fspace, 1)
    array = backend.array_constructor([[[-1, -3],
                       [2, 0]]], device=device)
    pwinner = PointwiseInner(vfspace, vecfield=array)

    testarr = backend.array_constructor([[1 + 1j, 2],
                        [3, 4 - 2j]], device=device)

    true_inner_adj = testarr[None, :, :] * array

    testfunc = fspace.element(testarr)
    testfunc_pwinner_adj = pwinner.adjoint(testfunc)
    assert all_almost_equal(testfunc_pwinner_adj, true_inner_adj)

    out = vfspace.element()
    pwinner.adjoint(testfunc, out=out)
    assert all_almost_equal(out, true_inner_adj)

    # 3d
    fspace = odl.uniform_discr([0, 0], [1, 1], (2, 2), dtype=complex, impl=impl, device=device)
    vfspace = ProductSpace(fspace, 3)
    array = backend.array_constructor([[[-1 - 1j, -3],
                       [2, 2j]],
                      [[-1j, 0],
                       [0, 1]],
                      [[-1, 1 + 2j],
                       [1, 1]]], device=device)
    pwinner = PointwiseInner(vfspace, vecfield=array)

    testarr = backend.array_constructor([[1 + 1j, 2],
                        [3, 4 - 2j]], device=device)

    true_inner_adj = testarr[None, :, :] * array

    testfunc = fspace.element(testarr)
    testfunc_pwinner_adj = pwinner.adjoint(testfunc)
    assert all_almost_equal(testfunc_pwinner_adj, true_inner_adj)

    out = vfspace.element()
    pwinner.adjoint(testfunc, out=out)
    assert all_almost_equal(out, true_inner_adj)


def test_pointwise_inner_adjoint_weighted(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    # Weighted product space only
    fspace = odl.uniform_discr([0, 0], [1, 1], (2, 2), dtype=complex, impl=impl, device=device)
    backend = fspace.array_backend
    vfspace = ProductSpace(fspace, 3, weighting=[2, 4, 6])
    array = backend.array_constructor([[[-1 - 1j, -3],
                       [2, 2j]],
                      [[-1j, 0],
                       [0, 1]],
                      [[-1, 1 + 2j],
                       [1, 1]]], device=device)
    pwinner = PointwiseInner(vfspace, vecfield=array)

    testarr = backend.array_constructor([[1 + 1j, 2],
                        [3, 4 - 2j]], device=device)

    true_inner_adj = testarr[None, :, :] * array  # same as unweighted case

    testfunc = fspace.element(testarr)
    testfunc_pwinner_adj = pwinner.adjoint(testfunc)
    assert all_almost_equal(testfunc_pwinner_adj, true_inner_adj)

    out = vfspace.element()
    pwinner.adjoint(testfunc, out=out)
    assert all_almost_equal(out, true_inner_adj)

    # Using different weighting in the inner product
    pwinner = PointwiseInner(vfspace, vecfield=array, weighting=[4, 8, 12])

    testarr = backend.array_constructor([[1 + 1j, 2],
                        [3, 4 - 2j]], device=device)

    true_inner_adj = 2 * testarr[None, :, :] * array  # w / v = (2, 2, 2)

    testfunc = fspace.element(testarr)
    testfunc_pwinner_adj = pwinner.adjoint(testfunc)
    assert all_almost_equal(testfunc_pwinner_adj, true_inner_adj)

    out = vfspace.element()
    pwinner.adjoint(testfunc, out=out)
    assert all_almost_equal(out, true_inner_adj)


# ---- PointwiseSum ----


def test_pointwise_sum(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    """PointwiseSum currently depends on PointwiseInner, we verify that."""

    fspace = odl.uniform_discr([0, 0], [1, 1], (2, 2), impl=impl, device=device)
    vfspace = ProductSpace(fspace, 3, exponent=2)

    # Make sure the code runs and test the properties
    psum = PointwiseSum(vfspace)
    assert isinstance(psum, PointwiseInner)
    assert psum.base_space == fspace
    assert all_equal(psum.weights, [1, 1, 1])
    assert all_equal(psum.vecfield, psum.domain.one())


# ---- MatrixOperator ---- #
def sparse_scipy_input(sparse_matrix_format):
    dense_matrix  = np.ones((3, 4))
    if sparse_matrix_format == 'COO':
        sparse_matrix = SparseMatrix('COO', 'scipy', dense_matrix)
    else:
        raise NotImplementedError
    return dense_matrix, sparse_matrix
    
def sparse_pytorch_input(sparse_matrix_format, cuda_device):
    assert sparse_matrix_format == 'COO', NotImplementedError
    indices = [
        #1st row|2nd row|3rd row
        [0,0,0,0,1,1,1,1,2,2,2,2],
        [0,1,2,3,0,1,2,3,0,1,2,3]
        ]
    values = [
        1.0,1.0,1.0,1.0,
        1.0,1.0,1.0,1.0,
        1.0,1.0,1.0,1.0
        ]
    array = [
        [1.0,1.0,1.0,1.0],
        [1.0,1.0,1.0,1.0],
        [1.0,1.0,1.0,1.0]
        ]
    backend = lookup_array_backend('pytorch')
    dense_matrix = backend.array_constructor(array, device=cuda_device)
    sparse_matrix = SparseMatrix('COO', 'pytorch', indices, values, device=cuda_device)
    return dense_matrix, sparse_matrix   



sparse_configs = []
sparse_configs.extend(
    (pytest.param(proj_cfg)
     for proj_cfg in ['COO scipy cpu'])
)

if 'pytorch' in tensor_space_impl_names():
    pytorch_cfgs = []
    for device in lookup_array_backend('pytorch').available_devices:
        pytorch_cfgs.append(f'COO pytorch {device}')

    sparse_configs.extend(
        (pytest.param(proj_cfg, marks=skip_if_no_pytorch)
         for proj_cfg in pytorch_cfgs)
    )

sparse_ids = [
    " format='{}' - backend='{}' - device='{}' ".format(*s.values[0].split())
    for s in sparse_configs
]

@pytest.fixture(scope='module', params=sparse_configs, ids=sparse_ids)
def matrix_input(request):
    format, backend, device = request.param.split()
    if backend == 'scipy':
        return sparse_scipy_input(format)
    elif backend == 'pytorch':
        return sparse_pytorch_input(format, device)
    else:
        raise ValueError
    
def invertible_sparse_scipy_input(sparse_matrix_format):
    assert sparse_matrix_format == 'COO', NotImplementedError
    dense_matrix  = np.ones((3, 3)) + 4.0 * np.eye(3)  # invertible
    sparse_matrix = SparseMatrix('COO', 'scipy', dense_matrix)
    return dense_matrix, sparse_matrix

def invertible_sparse_pytorch_input(sparse_matrix_format, cuda_device):
    assert sparse_matrix_format == 'COO', NotImplementedError
    indices = [
        #1st row|2nd row|3rd row
        [0,0,0,1,1,1,2,2,2],
        [0,1,2,0,1,2,0,1,2]
        ]
    values = [
        5.0,1.0,1.0,
        1.0,5.0,1.0,
        1.0,1.0,5.0
        ]
    array = [
        [5.0,1.0,1.0],
        [1.0,5.0,1.0],
        [1.0,1.0,5.0]
        ]
    backend = lookup_array_backend('pytorch')
    dense_matrix = backend.array_constructor(array, device=cuda_device)
    sparse_matrix = SparseMatrix('COO', 'pytorch', indices, values, device=cuda_device)
    return dense_matrix, sparse_matrix   
    
@pytest.fixture(scope='module', params=sparse_configs, ids=sparse_ids)
def invertible_matrix_input(request):
    format, backend, device = request.param.split()
    if backend == 'scipy':
        return invertible_sparse_scipy_input(format)
    elif backend == 'pytorch':
        return invertible_sparse_pytorch_input(format, device)
    else:
        raise ValueError

def test_matrix_op_init(matrix_input):
    """Test initialization and properties of matrix operators."""
    dense_matrix, sparse_matrix = matrix_input

    dense_matrix, backend = get_array_and_backend(dense_matrix)
    impl = backend.impl
    device = dense_matrix.device
    # Just check if the code runs
    MatrixOperator(dense_matrix)
    MatrixOperator(sparse_matrix)

    # Test default domain and range
    mat_op = MatrixOperator(dense_matrix)
    assert mat_op.domain == odl.tensor_space(4, dense_matrix.dtype, impl=impl, device=device)
    assert mat_op.range == odl.tensor_space(3, dense_matrix.dtype, impl=impl, device=device)
    assert odl.all(mat_op.matrix == dense_matrix)

    mat_op = MatrixOperator(sparse_matrix)
    assert mat_op.domain == odl.tensor_space(4, dense_matrix.dtype, impl=impl, device=device)
    assert mat_op.range == odl.tensor_space(3, dense_matrix.dtype, impl=impl, device=device)
    if impl == 'numpy':
        assert (mat_op.matrix != sparse_matrix).getnnz() == 0
    # Pytorch does not support == and != betweend sparse tensors
    elif impl == 'pytorch':
        assert len(mat_op.matrix) == len(sparse_matrix) 
    else:
        raise NotImplementedError
    # Explicit domain and range
    dom = odl.tensor_space(4, dense_matrix.dtype, impl=impl, device=device)
    ran = odl.tensor_space(3, dense_matrix.dtype, impl=impl, device=device)

    mat_op = MatrixOperator(dense_matrix, domain=dom, range=ran)
    assert mat_op.domain == dom
    assert mat_op.range == ran

    mat_op = MatrixOperator(sparse_matrix, domain=dom, range=ran)
    assert mat_op.domain == dom
    assert mat_op.range == ran

    # Bad 1d sizes
    with pytest.raises(ValueError):
        MatrixOperator(dense_matrix, domain=odl.cn(4, impl=impl, device=device), range=odl.cn(4, impl=impl, device=device))
    with pytest.raises(ValueError):
        MatrixOperator(dense_matrix, range=odl.cn(4, impl=impl, device=device))
    # Invalid range dtype
    with pytest.raises(ValueError):
        if impl == 'numpy':
            MatrixOperator(dense_matrix.astype(complex), range=odl.rn(4, impl=impl, device=device))
        elif impl == 'pytorch':
            MatrixOperator(dense_matrix.to(complex), range=odl.rn(4, impl=impl, device=device))
        else:
            raise NotImplementedError

    # Data type promotion
    # real space, complex matrix -> complex space
    dom = odl.rn(4, impl=impl, device=device)
    if impl == 'numpy':
        mat_op = MatrixOperator(dense_matrix.astype(complex), domain=dom, impl=impl, device=device)

    elif impl == 'pytorch':
        mat_op = MatrixOperator(dense_matrix.to(complex), domain=dom,
        impl=impl, device=device)
    else:
        raise NotImplementedError
    assert mat_op.domain == dom
    assert mat_op.range == odl.cn(3, impl=impl, device=device)

    # complex space, real matrix -> complex space
    dom = odl.cn(4, impl=impl, device=device)
    mat_op = MatrixOperator(dense_matrix.real, domain=dom)
    assert mat_op.domain == dom
    assert mat_op.range == odl.cn(3, impl=impl, device=device)

    # Multi-dimensional spaces
    dom = odl.tensor_space((6, 5, 4), dense_matrix.dtype, impl=impl, device=device)
    ran = odl.tensor_space((6, 5, 3), dense_matrix.dtype, impl=impl, device=device)
    mat_op = MatrixOperator(dense_matrix, domain=dom, axis=2)
    assert mat_op.range == ran
    mat_op = MatrixOperator(dense_matrix, domain=dom, range=ran, axis=2)
    assert mat_op.range == ran

    with pytest.raises(ValueError):
        bad_dom = odl.tensor_space((6, 6, 6), dense_matrix.dtype)  # wrong shape
        MatrixOperator(dense_matrix, domain=bad_dom)
    with pytest.raises(ValueError):
        dom = odl.tensor_space((6, 5, 4), dense_matrix.dtype)
        bad_ran = odl.tensor_space((6, 6, 6), dense_matrix.dtype)  # wrong shape
        MatrixOperator(dense_matrix, domain=dom, range=bad_ran)
    with pytest.raises(ValueError):
        MatrixOperator(dense_matrix, domain=dom, axis=1)
    with pytest.raises(ValueError):
        MatrixOperator(dense_matrix, domain=dom, axis=0)
    with pytest.raises(ValueError):
        bad_ran = odl.tensor_space((6, 3, 4), dense_matrix.dtype, impl=impl, device=device)
        MatrixOperator(dense_matrix, domain=dom, range=bad_ran, axis=2)
    with pytest.raises(ValueError):
        bad_dom_for_sparse = odl.rn((6, 5, 4), impl=impl, device=device)
        MatrixOperator(sparse_matrix, domain=bad_dom_for_sparse, axis=2, impl=impl, device=device)

    # Init with uniform_discr space (subclass of TensorSpace)
    dom = odl.uniform_discr(0, 1, 4, dtype=dense_matrix.dtype, impl=impl, device=device)
    ran = odl.uniform_discr(0, 1, 3, dtype=dense_matrix.dtype, impl=impl, device=device)
    MatrixOperator(dense_matrix, domain=dom, range=ran)

    # Make sure this runs and returns something string-like
    assert str(mat_op) > ''
    assert repr(mat_op) > ''


def test_matrix_op_call_implicit(matrix_input):
    """Validate result from calls to matrix operators against Numpy."""
    dense_matrix, sparse_matrix = matrix_input

    dense_matrix, backend = get_array_and_backend(dense_matrix)
    impl = backend.impl
    device = dense_matrix.device
    ns = backend.array_namespace

    # Default 1d case
    dmat_op = MatrixOperator(dense_matrix)
    smat_op = MatrixOperator(sparse_matrix)
    xarr, x = noise_elements(dmat_op.domain)
    # if impl == 'numpy':
    #     true_result = dense_matrix.dot(xarr)
    # elif impl == 'pytorch':
    
    true_result = ns.tensordot(dense_matrix, xarr, axes=([1], [0]))
    assert all_almost_equal(dmat_op(x), true_result)
    assert all_almost_equal(smat_op(x), true_result)
    out = dmat_op.range.element()
    dmat_op(x, out=out)
    assert all_almost_equal(out, true_result)
    smat_op(x, out=out)
    assert all_almost_equal(out, true_result)

    # Multi-dimensional case
    

    domain = odl.rn((2, 2, 4),impl=impl,device=device)
    mat_op = MatrixOperator(dense_matrix, domain, axis=2)
    xarr, x = noise_elements(mat_op.domain)
    true_result = ns.moveaxis(ns.tensordot(dense_matrix, xarr, axes=([1], [2])), 0, 2)
    assert all_almost_equal(mat_op(x), true_result)
    out = mat_op.range.element()
    mat_op(x, out=out)
    assert all_almost_equal(out, true_result)


def test_matrix_op_call_explicit(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    """Validate result from call to matrix op against explicit calculation."""

    space = odl.rn((3,2), impl=impl, device=device)
    mat = space.one().data

    backend = space.array_backend
    ns = space.array_namespace

    xarr = backend.array_constructor([[[0, 1],
                      [2, 3]],
                     [[4, 5],
                      [6, 7]]], dtype=float, device=device)

    # Multiplication along `axis` with `mat` is the same as summation
    # along `axis` and stacking 3 times along the same axis
    for axis in range(3):
        mat_op = MatrixOperator(mat, domain=odl.rn(xarr.shape, impl=impl, device=device),
                                axis=axis)
        result = mat_op(xarr)
        if impl == 'numpy':
            true_result = ns.repeat(ns.sum(xarr, axis=axis, keepdims=True),
                                repeats=3, axis=axis)
        elif impl == 'pytorch':
            true_result = ns.repeat_interleave(ns.sum(xarr, axis=axis, keepdims=True),
                                repeats=3, axis=axis)
        else:
            raise ValueError(f'Not implemented for impl = {impl}')
        assert result.shape == true_result.shape
        assert odl.allclose(result, true_result)


def test_matrix_op_adjoint(matrix_input):
    """Test if the adjoint of matrix operators is correct."""
    dense_matrix, sparse_matrix = matrix_input

    dense_matrix, backend = get_array_and_backend(dense_matrix)
    impl = backend.impl
    device = dense_matrix.device
    ns = backend.array_namespace
    tol = 2 * len(dense_matrix) * ns.finfo(dense_matrix.dtype).resolution
    # Default 1d case
    dmat_op = MatrixOperator(dense_matrix)
    smat_op = MatrixOperator(sparse_matrix)
    x = noise_element(dmat_op.domain)
    y = noise_element(dmat_op.range)

    inner_ran = dmat_op(x).inner(y)
    inner_dom = x.inner(dmat_op.adjoint(y))
    assert inner_ran == pytest.approx(inner_dom, rel=tol, abs=tol)
    inner_ran = smat_op(x).inner(y)
    inner_dom = x.inner(smat_op.adjoint(y))
    assert inner_ran == pytest.approx(inner_dom, rel=tol, abs=tol)

    # Multi-dimensional case
    domain = odl.tensor_space((2, 2, 4), impl=impl, device=device)
    mat_op = MatrixOperator(dense_matrix, domain, axis=2, impl=impl, device=device)
    x = noise_element(mat_op.domain)
    y = noise_element(mat_op.range)
    inner_ran = mat_op(x).inner(y)
    inner_dom = x.inner(mat_op.adjoint(y))
    assert inner_ran == pytest.approx(inner_dom, rel=tol, abs=tol)


def test_matrix_op_inverse(invertible_matrix_input):
    """Test if the inverse of matrix operators is correct."""
    dense_matrix, sparse_matrix = invertible_matrix_input

    # Default 1d case
    dmat_op = MatrixOperator(dense_matrix)
    smat_op = MatrixOperator(sparse_matrix)
    x = noise_element(dmat_op.domain)
    md_x = dmat_op(x)
    mdinv_md_x = dmat_op.inverse(md_x)
    assert all_almost_equal(x, mdinv_md_x)
    ms_x = smat_op(x)
    msinv_ms_x = smat_op.inverse(ms_x)
    assert all_almost_equal(x, msinv_ms_x)

    # Multi-dimensional case
    dense_matrix, backend = get_array_and_backend(dense_matrix)
    impl = backend.impl
    device = dense_matrix.device
    domain = odl.tensor_space((2, 2, 3), impl=impl, device=device)
    mat_op = MatrixOperator(dense_matrix, domain, axis=2)
    x = noise_element(mat_op.domain)
    m_x = mat_op(x)
    minv_m_x = mat_op.inverse(m_x)
    assert all_almost_equal(x, minv_m_x)


def test_sampling_operator_adjoint(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    """Validate basic properties of `SamplingOperator.adjoint`."""
    # 1d space
    space = odl.uniform_discr([-1], [1], shape=(3), impl=impl, device=device)
    sampling_points = [[0, 1, 1, 0]]
    x = space.element([1, 2, 3])
    op = odl.SamplingOperator(space, sampling_points)
    assert op.adjoint(op(x)).inner(x) == pytest.approx(op(x).inner(op(x)))

    op = odl.SamplingOperator(space, sampling_points, variant='integrate')
    assert op.adjoint(op(x)).inner(x) == pytest.approx(op(x).inner(op(x)))

    # 2d space
    space = odl.uniform_discr([-1, -1], [1, 1], shape=(2, 3), impl=impl, device=device)
    x = space.element([[1, 2, 3],
                       [4, 5, 6]])
    sampling_points = [[0, 1, 1, 0],
                       [0, 1, 2, 0]]
    op = odl.SamplingOperator(space, sampling_points)
    assert op.adjoint(op(x)).inner(x) == pytest.approx(op(x).inner(op(x)))

    # The ``'integrate'`` variant adjoint puts ones at the indices in
    # `sampling_points``, multiplied by their multiplicity:
    op = odl.SamplingOperator(space, sampling_points, variant='integrate')
    assert op.adjoint(op(x)).inner(x) == pytest.approx(op(x).inner(op(x)))


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Default operators defined on any `ProductSpace`."""

from __future__ import print_function, division, absolute_import
from numbers import Integral
import numpy as np

from odl.core.operator.operator import Operator
from odl.core.operator.default_ops import ZeroOperator
from odl.core.space import ProductSpace
from odl.core.util import COOMatrix


__all__ = ('ProductSpaceOperator',
           'ComponentProjection', 'ComponentProjectionAdjoint',
           'BroadcastOperator', 'ReductionOperator', 'DiagonalOperator')


class ProductSpaceOperator(Operator):

    r"""A "matrix of operators" on product spaces.

    For example a matrix of operators can act on a vector by

        ``ProductSpaceOperator([[A, B], [C, D]])([x, y]) =
        [A(x) + B(y), C(x) + D(y)]``

    Notes
    -----
    This is intended for the case where an operator can be decomposed
    as a linear combination of "sub-operators", e.g.

    .. math::
        \left(
        \begin{array}{ccc}
        A & B & 0 \\
        0 & C & 0 \\
        0 & 0 & D
        \end{array}\right)
        \left(
        \begin{array}{c}
        x \\
        y \\
        z
        \end{array}\right)
        =
        \left(
        \begin{array}{c}
        A(x) + B(y) \\
        C(y) \\
        D(z)
        \end{array}\right)

    Mathematically, a `ProductSpaceOperator` is an operator

    .. math::
        \mathcal{A}: \mathcal{X} \to \mathcal{Y}

    between product spaces
    :math:`\mathcal{X}=\mathcal{X}_1 \times\dots\times \mathcal{X}_m`
    and
    :math:`\mathcal{Y}=\mathcal{Y}_1 \times\dots\times \mathcal{Y}_n`
    which can be written in the form

    .. math::
        \mathcal{A} = (\mathcal{A}_{ij})_{i,j},  \quad
                          i = 1, \dots, n, \ j = 1, \dots, m

    with *component operators*
    :math:`\mathcal{A}_{ij}: \mathcal{X}_j \to \mathcal{Y}_i`.

    Its action on a vector :math:`x = (x_1, \dots, x_m)` is defined as
    the matrix multiplication

    .. math::
        [\mathcal{A}(x)]_i = \sum_{j=1}^m \mathcal{A}_{ij}(x_j).

    See Also
    --------
    BroadcastOperator : Case when a single argument is used by several ops.
    ReductionOperator : Calculates sum of operator results.
    DiagonalOperator : Case where the 'matrix' is diagonal.
    """

    def __init__(self, operators, domain=None, range=None):
        """Initialize a new instance.

        Parameters
        ----------
        operators : `array-like`
            An array of `Operator`'s, must be 2-dimensional.
        domain : `ProductSpace`, optional
            Domain of the operator. If not provided, it is tried to be
            inferred from the operators. This requires each **column**
            to contain at least one operator.
        range : `ProductSpace`, optional
            Range of the operator. If not provided, it is tried to be
            inferred from the operators. This requires each **row**
            to contain at least one operator.

        Examples
        --------
        >>> r3 = odl.rn(3)
        >>> pspace = odl.ProductSpace(r3, r3)
        >>> I = odl.IdentityOperator(r3)
        >>> x = pspace.element([[1, 2, 3],
        ...                     [4, 5, 6]])

        Create an operator that sums two inputs:

        >>> prod_op = odl.ProductSpaceOperator([[I, I]])
        >>> prod_op(x)
        ProductSpace(rn(3), 1).element([
            [ 5.,  7.,  9.]
        ])

        Diagonal operator -- 0 or ``None`` means ignore, or the implicit
        zero operator:

        >>> prod_op = odl.ProductSpaceOperator([[I, 0],
        ...                                     [0, I]])
        >>> prod_op(x)
        ProductSpace(rn(3), 2).element([
            [ 1.,  2.,  3.],
            [ 4.,  5.,  6.]
        ])

        If a column is empty, the operator domain must be specified. The
        same holds for an empty row and the range of the operator:

        >>> prod_op = odl.ProductSpaceOperator([[I, 0],
        ...                                     [I, 0]], domain=r3 ** 2)
        >>> prod_op(x)
        ProductSpace(rn(3), 2).element([
            [ 1.,  2.,  3.],
            [ 1.,  2.,  3.]
        ])
        >>> prod_op = odl.ProductSpaceOperator([[I, I],
        ...                                     [0, 0]], range=r3 ** 2)
        >>> prod_op(x)
        ProductSpace(rn(3), 2).element([
            [ 5.,  7.,  9.],
            [ 0.,  0.,  0.]
        ])
        """
        # Lazy import to improve `import odl` time
        import scipy.sparse

        # Validate input data
        if domain is not None:
            if not isinstance(domain, ProductSpace):
                raise TypeError('`domain` {!r} not a ProductSpace instance'
                                ''.format(domain))
            if domain.is_weighted:
                raise NotImplementedError('weighted spaces not supported')

        if range is not None:
            if not isinstance(range, ProductSpace):
                raise TypeError('`range` {!r} not a ProductSpace instance'
                                ''.format(range))
            if range.is_weighted:
                raise NotImplementedError('weighted spaces not supported')

        if isinstance(operators, scipy.sparse.spmatrix):
            if not all(isinstance(op, Operator) for op in operators.data):
                raise ValueError('sparse matrix `operator` contains non-'
                                 '`Operator` entries')
            # scipy sparse matrix not supported (deprecated due to API changes)
            # keep now for backward compatibility
            print('Warning: scipy.sparse.spmatrix is deprecated.')
            self.__ops = COOMatrix(operators.data,
                                   (operators.row, operators.col),
                                   operators.shape)

        elif isinstance(operators, COOMatrix):
            self.__ops = operators

        else:
            self.__ops = self._convert_to_spmatrix(operators)

        # Set domain and range (or verify if given)
        if domain is None:
            domains = [None] * self.__ops.shape[1]
        else:
            domains = domain

        if range is None:
            ranges = [None] * self.__ops.shape[0]
        else:
            ranges = range

        for row, col, op in zip(self.__ops.row, self.__ops.col,
                                self.__ops.data):
            if domains[col] is None:
                domains[col] = op.domain
            elif domains[col] != op.domain:
                raise ValueError('column {}, has inconsistent domains, '
                                 'got {} and {}'
                                 ''.format(col, domains[col], op.domain))

            if ranges[row] is None:
                ranges[row] = op.range
            elif ranges[row] != op.range:
                raise ValueError('row {}, has inconsistent ranges, '
                                 'got {} and {}'
                                 ''.format(row, ranges[row], op.range))

        if domain is None:
            for col, sub_domain in enumerate(domains):
                if sub_domain is None:
                    raise ValueError('col {} empty, unable to determine '
                                     'domain, please use `domain` parameter'
                                     ''.format(col))

            domain = ProductSpace(*domains)

        if range is None:
            for row, sub_range in enumerate(ranges):
                if sub_range is None:
                    raise ValueError('row {} empty, unable to determine '
                                     'range, please use `range` parameter'
                                     ''.format(row))

            range = ProductSpace(*ranges)

        # Set linearity
        linear = all(op.is_linear for op in self.__ops.data)

        super(ProductSpaceOperator, self).__init__(
            domain=domain, range=range, linear=linear)

    @staticmethod
    def _convert_to_spmatrix(operators):
        """Convert an array-like object of operators to a sparse matrix."""

        # Convert ops to sparse representation. This is not trivial because
        # operators can be indexable themselves and give the wrong impression
        # of an extra dimension. So we have to infer the shape manually
        # first and extract the indices of nonzero positions.
        nrows = len(operators)
        ncols = None
        irow, icol, data = [], [], []
        for i, row in enumerate(operators):
            try:
                iter(row)
            except TypeError:
                raise ValueError(
                    '`operators` must be a matrix of `Operator` objects, `0` '
                    'or `None`, got {!r} (row {} = {!r} is not iterable)'
                    ''.format(operators, i, row))

            if isinstance(row, Operator):
                raise ValueError(
                    '`operators` must be a matrix of `Operator` objects, `0` '
                    'or `None`, but row {} is an `Operator` {!r}'
                    ''.format(i, row))

            if ncols is None:
                ncols = len(row)
            elif len(row) != ncols:
                raise ValueError(
                    'all rows in `operators` must have the same length, but '
                    'length {} of row {} differs from previous common length '
                    '{}'.format(len(row), i, ncols))

            for j, col in enumerate(row):
                if col is None or col == 0:
                    pass
                elif isinstance(col, Operator):
                    irow.append(i)
                    icol.append(j)
                    data.append(col)
                else:
                    raise ValueError(
                        '`operators` must be a matrix of `Operator` objects, '
                        '`0` or `None`, got entry {!r} at ({}, {})'
                        ''.format(col, i, j))

        # Create object array explicitly, threby avoiding erroneous conversion
        # in `coo_matrix.__init__`
        data_arr = np.empty(len(data), dtype=object)
        data_arr[:] = data

        return COOMatrix(data_arr, (irow, icol), (nrows, ncols))

    @property
    def ops(self):
        """The sparse operator matrix representing this operator."""
        return self.__ops

    def _call(self, x, out=None):
        """Call the operators on the parts of ``x``."""
        # TODO: add optimization in case an operator appears repeatedly in a
        # row
        if out is None:
            out = self.range.zero()
            for i, j, op in zip(self.ops.row, self.ops.col, self.ops.data):
                out[i] += op(x[j])
        else:
            has_evaluated_row = np.zeros(len(self.range), dtype=bool)
            for i, j, op in zip(self.ops.row, self.ops.col, self.ops.data):
                if not has_evaluated_row[i]:
                    op(x[j], out=out[i])
                else:
                    # TODO: optimize
                    out[i] += op(x[j])

                has_evaluated_row[i] = True

            for i, evaluated in enumerate(has_evaluated_row):
                if not evaluated:
                    out[i].set_zero()

        return out

    def derivative(self, x):
        """Derivative of the product space operator.

        Parameters
        ----------
        x : `domain` element
            The point to take the derivative in

        Returns
        -------
        adjoint : linear`ProductSpaceOperator`
            The derivative

        Examples
        --------
        >>> r3 = odl.rn(3)
        >>> pspace = odl.ProductSpace(r3, r3)
        >>> I = odl.IdentityOperator(r3)
        >>> x = pspace.element([[1, 2, 3], [4, 5, 6]])

        Example with linear operator (derivative is itself)

        >>> prod_op = ProductSpaceOperator([[0, I], [0, 0]],
        ...                                domain=pspace, range=pspace)
        >>> prod_op(x)
        ProductSpace(rn(3), 2).element([
            [ 4.,  5.,  6.],
            [ 0.,  0.,  0.]
        ])
        >>> prod_op.derivative(x)(x)
        ProductSpace(rn(3), 2).element([
            [ 4.,  5.,  6.],
            [ 0.,  0.,  0.]
        ])

        Example with affine operator

        >>> residual_op = I - r3.element([1, 1, 1])
        >>> op = ProductSpaceOperator([[0, residual_op], [0, 0]],
        ...                           domain=pspace, range=pspace)

        Calling operator gives offset by [1, 1, 1]

        >>> op(x)
        ProductSpace(rn(3), 2).element([
            [ 3.,  4.,  5.],
            [ 0.,  0.,  0.]
        ])

        Derivative of affine operator does not have this offset

        >>> op.derivative(x)(x)
        ProductSpace(rn(3), 2).element([
            [ 4.,  5.,  6.],
            [ 0.,  0.,  0.]
        ])
        """
        # Short circuit optimization
        if self.is_linear:
            return self

        deriv_ops = [op.derivative(x[col]) for op, col in zip(self.ops.data,
                                                              self.ops.col)]
        data = np.empty(len(deriv_ops), dtype=object)
        data[:] = deriv_ops
        indices = [self.ops.row, self.ops.col]
        shape = self.ops.shape
        deriv_matrix = COOMatrix(data, indices, shape)
        return ProductSpaceOperator(deriv_matrix, self.domain, self.range)

    @property
    def adjoint(self):
        """Adjoint of this operator.

        The adjoint is given by taking the transpose of the matrix
        and the adjoint of each component operator.

        In weighted product spaces, the adjoint needs to take the
        weightings into account. This is currently not supported.

        Returns
        -------
        adjoint : `ProductSpaceOperator`
            The adjoint

        Examples
        --------
        >>> r3 = odl.rn(3)
        >>> pspace = odl.ProductSpace(r3, r3)
        >>> I = odl.IdentityOperator(r3)
        >>> x = pspace.element([[1, 2, 3],
        ...                     [4, 5, 6]])

        Matrix is transposed:

        >>> prod_op = ProductSpaceOperator([[0, I], [0, 0]],
        ...                                domain=pspace, range=pspace)
        >>> prod_op(x)
        ProductSpace(rn(3), 2).element([
            [ 4.,  5.,  6.],
            [ 0.,  0.,  0.]
        ])
        >>> prod_op.adjoint(x)
        ProductSpace(rn(3), 2).element([
            [ 0.,  0.,  0.],
            [ 1.,  2.,  3.]
        ])
        """
        adjoint_ops = [op.adjoint for op in self.ops.data]
        data = np.empty(len(adjoint_ops), dtype=object)
        data[:] = adjoint_ops
        indices = [self.ops.col, self.ops.row]  # Swap col/row -> transpose
        shape = (self.ops.shape[1], self.ops.shape[0])
        adj_matrix = COOMatrix(data, indices, shape)
        return ProductSpaceOperator(adj_matrix, self.range, self.domain)

    def __getitem__(self, index):
        """Get sub-operator by index.

        Parameters
        ----------
        index : int or tuple of int
            A pair of integers given as (row, col).

        Returns
        -------
        suboperator : `ReductionOperator`, `Operator` or ``0``
            If index is an integer, return the row given by the index.

            If index is a tuple, it must have two elements.
            if there is an operator at ``(row, col)``,  the operator is
            returned, otherwise ``0``.

        Examples
        --------
        >>> r3 = odl.rn(3)
        >>> pspace = odl.ProductSpace(r3, r3)
        >>> I = odl.IdentityOperator(r3)
        >>> prod_op = ProductSpaceOperator([[0, I],
        ...                                 [0, 0]],
        ...                                domain=pspace, range=pspace)
        >>> prod_op[0, 0]
        0
        >>> prod_op[0, 1]
        IdentityOperator(rn(3))
        >>> prod_op[1, 0]
        0
        >>> prod_op[1, 1]
        0

        By accessing single indices, a row is extracted as a
        `ReductionOperator`:

        >>> prod_op[0]
        ReductionOperator(ZeroOperator(rn(3)), IdentityOperator(rn(3)))
        """
        if isinstance(index, tuple):
            row, col = index

            linear_index = np.flatnonzero((self.ops.row == row) &
                                          (self.ops.col == col))
            if linear_index.size == 0:
                return 0
            else:
                return self.ops.data[int(linear_index)]
        else:
            index = int(index)

            ops = [None] * len(self.domain)
            for op, col, row in zip(self.ops.data, self.ops.col, self.ops.row):
                if row == index:
                    ops[col] = op

            for i in range(len(self.domain)):
                if ops[i] is None:
                    ops[i] = ZeroOperator(self.domain[i])

            return ReductionOperator(*ops)

    @property
    def shape(self):
        """Shape of the matrix of operators."""
        return self.ops.shape

    def __len__(self):
        """Return ``len(self)``."""
        return self.shape[0]

    @property
    def size(self):
        """Total size of the matrix of operators."""
        return np.prod(self.shape, dtype='int64')

    def __repr__(self):
        """Return ``repr(self)``."""
        aslist = [[0] * len(self.domain) for _ in range(len(self.range))]
        for i, j, op in zip(self.ops.row, self.ops.col, self.ops.data):
            aslist[i][j] = op
        return '{}({!r})'.format(self.__class__.__name__, aslist)


class ComponentProjection(Operator):

    r"""Projection onto the subspace identified by an index.

    For a product space :math:`\mathcal{X} = \mathcal{X}_1 \times \dots
    \times \mathcal{X}_n`, the component projection

    .. math::
       \mathcal{P}_i: \mathcal{X} \to \mathcal{X}_i

    is given by :math:`\mathcal{P}_i(x) = x_i` for an element
    :math:`x = (x_1, \dots, x_n) \in \mathcal{X}`.

    More generally, for an index set :math:`I \subset \{1, \dots, n\}`,
    the projection operator :math:`\mathcal{P}_I` is defined by
    :math:`\mathcal{P}_I(x) = (x_i)_{i \in I}`.

    Note that this is a special case of a product space operator where
    the "operator matrix" has only one row and contains only
    identity operators.
    """

    def __init__(self, space, index):
        """Initialize a new instance.

        Parameters
        ----------
        space : `ProductSpace`
            Space to project from.
        index : int, slice, or list
            Indices defining the subspace. If ``index`` is not an integer,
            the `Operator.range` of this operator is also a `ProductSpace`.

        Examples
        --------
        >>> r1 = odl.rn(1)
        >>> r2 = odl.rn(2)
        >>> r3 = odl.rn(3)
        >>> pspace = odl.ProductSpace(r1, r2, r3)

        Projection on n-th component:

        >>> proj = odl.ComponentProjection(pspace, 0)
        >>> x = [[1],
        ...      [2, 3],
        ...      [4, 5, 6]]
        >>> proj(x)
        rn(1).element([ 1.])

        Projection on sub-space:

        >>> proj = odl.ComponentProjection(pspace, [0, 2])
        >>> proj(x)
        ProductSpace(rn(1), rn(3)).element([
            [ 1.],
            [ 4.,  5.,  6.]
        ])
        """
        self.__index = index
        super(ComponentProjection, self).__init__(
            space, space[index], linear=True)

    @property
    def index(self):
        """Index of the subspace."""
        return self.__index

    def _call(self, x, out=None):
        """Project ``x`` onto the subspace."""
        if out is None:
            out = x[self.index].copy()
        else:
            out.assign(x[self.index])
        return out

    @property
    def adjoint(self):
        """The adjoint operator.

        The adjoint is given by extending along `ComponentProjection.index`,
        and setting zero along the others.

        See Also
        --------
        ComponentProjectionAdjoint
        """
        return ComponentProjectionAdjoint(self.domain, self.index)

    def __repr__(self):
        """Return ``repr(self)``.

        Examples
        --------
        >>> pspace = odl.ProductSpace(odl.rn(1), odl.rn(2))
        >>> odl.ComponentProjection(pspace, 0)
        ComponentProjection(ProductSpace(rn(1), rn(2)), 0)
        """
        return '{}({!r}, {})'.format(self.__class__.__name__,
                                     self.domain, self.index)


class ComponentProjectionAdjoint(Operator):

    """Adjoint operator to `ComponentProjection`.

    As a special case of the adjoint of a `ProductSpaceOperator`,
    this operator is given as a column vector of identity operators
    and zero operators, with the identities placed in the positions
    defined by `ComponentProjectionAdjoint.index`.

    In weighted product spaces, the adjoint needs to take the
    weightings into account. This is currently not supported.
    """

    def __init__(self, space, index):
        """Initialize a new instance

        Parameters
        ----------
        space : `ProductSpace`
            Space to project to.
        index : int, slice, or list
            Indexes to project from.

        Examples
        --------
        >>> r1 = odl.rn(1)
        >>> r2 = odl.rn(2)
        >>> r3 = odl.rn(3)
        >>> pspace = odl.ProductSpace(r1, r2, r3)
        >>> x = pspace.element([[1],
        ...                     [2, 3],
        ...                     [4, 5, 6]])

        Projection on the 0-th component:

        >>> proj_adj = odl.ComponentProjectionAdjoint(pspace, 0)
        >>> proj_adj(x[0])
        ProductSpace(rn(1), rn(2), rn(3)).element([
            [ 1.],
            [ 0.,  0.],
            [ 0.,  0.,  0.]
        ])

        Projection on a sub-space corresponding to indices 0 and 2:

        >>> proj_adj = odl.ComponentProjectionAdjoint(pspace, [0, 2])
        >>> proj_adj(x[[0, 2]])
        ProductSpace(rn(1), rn(2), rn(3)).element([
            [ 1.],
            [ 0.,  0.],
            [ 4.,  5.,  6.]
        ])
        """
        self.__index = index
        super(ComponentProjectionAdjoint, self).__init__(
            space[index], space, linear=True)

    @property
    def index(self):
        """Index of the subspace."""
        return self.__index

    def _call(self, x, out=None):
        """Extend ``x`` from the subspace."""
        if out is None:
            out = self.range.zero()
        else:
            out.set_zero()

        out[self.index] = x
        return out

    @property
    def adjoint(self):
        """Adjoint of this operator.

        Returns
        -------
        adjoint : `ComponentProjection`
            The adjoint is given by the `ComponentProjection` related to this
            operator's `index`.
        """
        return ComponentProjection(self.range, self.index)

    def __repr__(self):
        """Return ``repr(self)``.

        Examples
        --------
        >>> pspace = odl.ProductSpace(odl.rn(1), odl.rn(2))
        >>> odl.ComponentProjectionAdjoint(pspace, 0)
        ComponentProjectionAdjoint(ProductSpace(rn(1), rn(2)), 0)
        """
        return '{}({!r}, {})'.format(self.__class__.__name__,
                                     self.range, self.index)


class BroadcastOperator(Operator):
    """Broadcast argument to set of operators.

    An argument is broadcast by evaluating several operators in the same
    point::

        BroadcastOperator(op1, op2)(x) = [op1(x), op2(x)]

    See Also
    --------
    ProductSpaceOperator : More general case, used as backend.
    ReductionOperator : Calculates sum of operator results.
    DiagonalOperator : Case where each operator should have its own argument.
    """

    def __init__(self, *operators):
        """Initialize a new instance

        Parameters
        ----------
        operator1,...,operatorN : `Operator` or `int`
            The individual operators that should be evaluated.
            Can also be given as ``operator, n`` with ``n`` integer,
            in which case ``operator`` is repeated ``n`` times.

        Examples
        --------
        Initialize an operator:

        >>> I = odl.IdentityOperator(odl.rn(3))
        >>> op = BroadcastOperator(I, 2 * I)
        >>> op.domain
        rn(3)
        >>> op.range
        ProductSpace(rn(3), 2)

        Evaluate the operator:

        >>> x = [1, 2, 3]
        >>> op(x)
        ProductSpace(rn(3), 2).element([
            [ 1.,  2.,  3.],
            [ 2.,  4.,  6.]
        ])

        Can also initialize by calling an operator repeatedly:

        >>> I = odl.IdentityOperator(odl.rn(3))
        >>> op = BroadcastOperator(I, 2)
        >>> op.operators
        (IdentityOperator(rn(3)), IdentityOperator(rn(3)))
        """
        if (len(operators) == 2 and
                isinstance(operators[0], Operator) and
                isinstance(operators[1], Integral)):
            operators = (operators[0],) * operators[1]

        self.__operators = operators
        self.__prod_op = ProductSpaceOperator([[op] for op in operators])
        super(BroadcastOperator, self).__init__(
            self.prod_op.domain[0], self.prod_op.range,
            linear=self.prod_op.is_linear)

    @property
    def prod_op(self):
        """`ProductSpaceOperator` implementation."""
        return self.__prod_op

    @property
    def operators(self):
        """Tuple of sub-operators that comprise ``self``."""
        return self.__operators

    def __getitem__(self, index):
        """Return ``self(index)``."""
        return self.operators[index]

    def __len__(self):
        """Return ``len(self)``."""
        return len(self.operators)

    @property
    def size(self):
        """Total number of sub-operators."""
        return len(self)

    def _call(self, x, out=None):
        """Evaluate all operators in ``x`` and broadcast."""
        wrapped_x = self.prod_op.domain.element([x], cast=False)
        return self.prod_op(wrapped_x, out=out)

    def derivative(self, x):
        """Derivative of the broadcast operator.

        Parameters
        ----------
        x : `domain` element
            The point to take the derivative in

        Returns
        -------
        adjoint : linear `BroadcastOperator`
            The derivative

        Examples
        --------
        Example with an affine operator:

        >>> I = odl.IdentityOperator(odl.rn(3))
        >>> residual_op = I - I.domain.element([1, 1, 1])
        >>> op = BroadcastOperator(residual_op, 2 * residual_op)

        Calling operator offsets by ``[1, 1, 1]``:

        >>> x = [1, 2, 3]
        >>> op(x)
        ProductSpace(rn(3), 2).element([
            [ 0.,  1.,  2.],
            [ 0.,  2.,  4.]
        ])

        The derivative of this affine operator does not have an offset:

        >>> op.derivative(x)(x)
        ProductSpace(rn(3), 2).element([
            [ 1.,  2.,  3.],
            [ 2.,  4.,  6.]
        ])
        """
        return BroadcastOperator(*[op.derivative(x) for op in
                                   self.operators])

    @property
    def adjoint(self):
        """Adjoint of this operator.

        Returns
        -------
        adjoint : linear `BroadcastOperator`

        Examples
        --------
        >>> I = odl.IdentityOperator(odl.rn(3))
        >>> op = BroadcastOperator(I, 2 * I)
        >>> op.adjoint([[1, 2, 3], [2, 3, 4]])
        rn(3).element([  5.,   8.,  11.])
        """
        return ReductionOperator(*[op.adjoint for op in self.operators])

    def __repr__(self):
        """Return ``repr(self)``.

        Examples
        --------
        >>> spc = odl.rn(3)
        >>> id = odl.IdentityOperator(spc)
        >>> odl.BroadcastOperator(id, 3)
        BroadcastOperator(IdentityOperator(rn(3)), 3)
        >>> scale = odl.ScalingOperator(spc, 3)
        >>> odl.BroadcastOperator(id, scale)
        BroadcastOperator(IdentityOperator(rn(3)), ScalingOperator(rn(3), 3.0))
        """
        if all(op == self[0] for op in self):
            return '{}({!r}, {})'.format(self.__class__.__name__,
                                         self[0], len(self))
        else:
            op_repr = ', '.join(repr(op) for op in self)
            return '{}({})'.format(self.__class__.__name__, op_repr)


class ReductionOperator(Operator):
    """Reduce argument over set of operators.

    An argument is reduced by evaluating several operators and summing the
    result::

        ReductionOperator(op1, op2)(x) = op1(x[0]) + op2(x[1])

    See Also
    --------
    ProductSpaceOperator : More general case, used as backend.
    BroadcastOperator : Calls several operators with same argument.
    DiagonalOperator : Case where each operator should have its own argument.
    SeparableSum : Corresponding construction for functionals.
    """

    def __init__(self, *operators):
        """Initialize a new instance.

        Parameters
        ----------
        operator1,...,operatorN : `Operator` or `int`
            The individual operators that should be evaluated and summed.
            Can also be given as ``operator, n`` with ``n`` integer,
            in which case ``operator`` is repeated ``n`` times.

        Examples
        --------
        >>> I = odl.IdentityOperator(odl.rn(3))
        >>> op = ReductionOperator(I, 2 * I)
        >>> op.domain
        ProductSpace(rn(3), 2)
        >>> op.range
        rn(3)

        Evaluating in a point gives the sum of the evaluation results of
        the individual operators:

        >>> op([[1, 2, 3],
        ...     [4, 6, 8]])
        rn(3).element([  9.,  14.,  19.])

        An ``out`` argument can be given for in-place evaluation:

        >>> out = op.range.element()
        >>> result = op([[1, 2, 3],
        ...              [4, 6, 8]], out=out)
        >>> out
        rn(3).element([  9.,  14.,  19.])
        >>> result is out
        True

        There is a simplified syntax for the case that all operators are
        the same:

        >>> op = ReductionOperator(I, 2)
        >>> op.operators
        (IdentityOperator(rn(3)), IdentityOperator(rn(3)))
        """
        if (len(operators) == 2 and
                isinstance(operators[0], Operator) and
                isinstance(operators[1], Integral)):
            operators = (operators[0],) * operators[1]

        self.__operators = operators
        self.__prod_op = ProductSpaceOperator([operators])

        super(ReductionOperator, self).__init__(
            self.prod_op.domain, self.prod_op.range[0],
            linear=self.prod_op.is_linear)

    @property
    def prod_op(self):
        """`ProductSpaceOperator` implementation."""
        return self.__prod_op

    @property
    def operators(self):
        """Tuple of sub-operators that comprise ``self``."""
        return self.__operators

    def __getitem__(self, index):
        """Return an operator by index."""
        return self.operators[index]

    def __len__(self):
        """Return ``len(self)``."""
        return len(self.operators)

    @property
    def size(self):
        """Total number of sub-operators."""
        return len(self)

    def _call(self, x, out=None):
        """Apply operators to ``x`` and sum."""
        if out is None:
            return self.prod_op(x)[0]
        else:
            wrapped_out = self.prod_op.range.element([out], cast=False)
            pspace_result = self.prod_op(x, out=wrapped_out)
            return pspace_result[0]

    def derivative(self, x):
        """Derivative of the reduction operator.

        Parameters
        ----------
        x : `domain` element
            The point to take the derivative in.

        Returns
        -------
        derivative : linear `BroadcastOperator`

        Examples
        --------
        >>> r3 = odl.rn(3)
        >>> I = odl.IdentityOperator(r3)
        >>> x = [1.0, 2.0, 3.0]
        >>> y = [4.0, 6.0, 8.0]

        Example with linear operator (derivative is itself)

        >>> op = ReductionOperator(I, 2 * I)
        >>> op([x, y])
        rn(3).element([  9.,  14.,  19.])
        >>> op.derivative([x, y])([x, y])
        rn(3).element([  9.,  14.,  19.])

        Example with affine operator

        >>> residual_op = I - r3.element([1, 1, 1])
        >>> op = ReductionOperator(residual_op, 2 * residual_op)

        Calling operator gives offset by [3, 3, 3]

        >>> op([x, y])
        rn(3).element([  6.,  11.,  16.])

        Derivative of affine operator does not have this offset

        >>> op.derivative([x, y])([x, y])
        rn(3).element([  9.,  14.,  19.])
        """
        return ReductionOperator(*[op.derivative(xi)
                                   for op, xi in zip(self.operators, x)])

    @property
    def adjoint(self):
        """Adjoint of this operator.

        Returns
        -------
        adjoint : linear `BroadcastOperator`

        Examples
        --------
        >>> I = odl.IdentityOperator(odl.rn(3))
        >>> op = ReductionOperator(I, 2 * I)
        >>> op.adjoint([1, 2, 3])
        ProductSpace(rn(3), 2).element([
            [ 1.,  2.,  3.],
            [ 2.,  4.,  6.]
        ])
        """
        return BroadcastOperator(*[op.adjoint for op in self.operators])

    def __repr__(self):
        """Return ``repr(self)``.

        Examples
        --------
        >>> spc = odl.rn(3)
        >>> id = odl.IdentityOperator(spc)
        >>> odl.ReductionOperator(id, 3)
        ReductionOperator(IdentityOperator(rn(3)), 3)
        >>> scale = odl.ScalingOperator(spc, 3)
        >>> odl.ReductionOperator(id, scale)
        ReductionOperator(IdentityOperator(rn(3)), ScalingOperator(rn(3), 3.0))
        """
        if all(op == self[0] for op in self):
            return '{}({!r}, {})'.format(self.__class__.__name__,
                                         self[0], len(self))
        else:
            op_repr = ', '.join(repr(op) for op in self)
            return '{}({})'.format(self.__class__.__name__, op_repr)


class DiagonalOperator(ProductSpaceOperator):
    """Diagonal 'matrix' of operators.

    For example, if ``A`` and ``B`` are operators, the diagonal operator
    can be seen as a matrix of operators::

        [[A, 0],
         [0, B]]

    When evaluated it gives::

         DiagonalOperator(op1, op2)(x) = [op1(x[0]), op2(x[1])]

    See Also
    --------
    ProductSpaceOperator : Case when the 'matrix' is dense.
    BroadcastOperator : Case when a single argument is used by several ops.
    ReductionOperator : Calculates sum of operator results.
    """

    def __init__(self, *operators, **kwargs):
        """Initialize a new instance.

        Parameters
        ----------
        operator1,...,operatorN : `Operator` or int
            The individual operators in the diagonal.
            Can be specified as ``operator, n`` with ``n`` integer,
            in which case the diagonal operator with ``n`` multiples of
            ``operator`` is created.
        kwargs :
            Keyword arguments passed to the `ProductSpaceOperator` backend.

        Examples
        --------
        >>> I = odl.IdentityOperator(odl.rn(3))
        >>> op = DiagonalOperator(I, 2 * I)
        >>> op.domain
        ProductSpace(rn(3), 2)
        >>> op.range
        ProductSpace(rn(3), 2)

        Evaluation is distributed so each argument is given to one operator.
        The argument order is the same as the order of the operators:

        >>> op([[1, 2, 3],
        ...     [4, 5, 6]])
        ProductSpace(rn(3), 2).element([
            [ 1.,  2.,  3.],
            [  8.,  10.,  12.]
        ])

        Can also be created using a multiple of a single operator

        >>> op = DiagonalOperator(I, 2)
        >>> op.operators
        (IdentityOperator(rn(3)), IdentityOperator(rn(3)))
        """

        if (len(operators) == 2 and
                isinstance(operators[0], Operator) and
                isinstance(operators[1], Integral)):
            operators = (operators[0],) * operators[1]

        n_ops = len(operators)
        irow = icol = np.arange(n_ops)
        data = np.empty(n_ops, dtype=object)
        data[:] = operators
        shape = (n_ops, n_ops)
        op_matrix = COOMatrix(data, (irow, icol), shape)
        super(DiagonalOperator, self).__init__(op_matrix, **kwargs)

        self.__operators = tuple(operators)

    @property
    def operators(self):
        """Tuple of sub-operators that comprise ``self``."""
        return self.__operators

    def __getitem__(self, index):
        """Return an operator by index."""
        return self.operators[index]

    def __len__(self):
        """Return ``len(self)``."""
        return len(self.operators)

    @property
    def size(self):
        """Total number of sub-operators."""
        return len(self)

    def derivative(self, point):
        """Derivative of this operator.

        For example, if A and B are operators

            [[A, 0],
             [0, B]]

        The derivative is given by:

            [[A', 0],
             [0, B']]

        This is only well defined if each sub-operator has a derivative

        Parameters
        ----------
        point : `element-like` in ``domain``
            The point in which the derivative should be taken.

        Returns
        -------
        derivative : `DiagonalOperator`
            The derivative operator

        See Also
        --------
        ProductSpaceOperator.derivative
        """
        point = self.domain.element(point)

        derivs = [op.derivative(p) for op, p in zip(self.operators, point)]
        return DiagonalOperator(*derivs,
                                domain=self.domain, range=self.range)

    @property
    def adjoint(self):
        """Adjoint of this operator.

        For example, if A and B are operators::

            [[A, 0],
             [0, B]]

        The adjoint is given by::

            [[A^*, 0],
             [0, B^*]]

        This is only well defined if each sub-operator has an adjoint

        Returns
        -------
        adjoint : `DiagonalOperator`
            The adjoint operator

        See Also
        --------
        ProductSpaceOperator.adjoint
        """
        adjoints = [op.adjoint for op in self.operators]
        return DiagonalOperator(*adjoints,
                                domain=self.range, range=self.domain)

    @property
    def inverse(self):
        """Inverse of this operator.

        For example, if A and B are operators::

            [[A, 0],
             [0, B]]

        The inverse is given by::

            [[A^-1, 0],
             [0, B^-1]]

        This is only well defined if each sub-operator has an inverse

        Returns
        -------
        inverse : `DiagonalOperator`
            The inverse operator

        See Also
        --------
        ProductSpaceOperator.inverse
        """
        inverses = [op.inverse for op in self.operators]
        return DiagonalOperator(*inverses,
                                domain=self.range, range=self.domain)

    def __repr__(self):
        """Return ``repr(self)``.

        Examples
        --------
        >>> spc = odl.rn(3)
        >>> id = odl.IdentityOperator(spc)
        >>> odl.DiagonalOperator(id, 3)
        DiagonalOperator(IdentityOperator(rn(3)), 3)
        >>> scale = odl.ScalingOperator(spc, 3)
        >>> odl.DiagonalOperator(id, scale)
        DiagonalOperator(IdentityOperator(rn(3)), ScalingOperator(rn(3), 3.0))
        """
        if all(op == self[0] for op in self):
            return '{}({!r}, {})'.format(self.__class__.__name__,
                                         self[0], len(self))
        else:
            op_repr = ', '.join(repr(op) for op in self)
            return '{}({})'.format(self.__class__.__name__, op_repr)


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Test configuration file."""

from __future__ import absolute_import, division, print_function

import operator
import os
from os import path

import numpy as np

import odl
from odl.core.array_API_support import lookup_array_backend
from odl.core.space.entry_points import tensor_space_impl_names
from odl.trafos.backends import PYFFTW_AVAILABLE, PYWT_AVAILABLE
from odl.core.util.testutils import simple_fixture
from odl.core.util.dtype_utils import INTEGER_DTYPES, FLOAT_DTYPES, COMPLEX_DTYPES

try:
    import pytest
    from pytest import fixture
except ImportError:
    pytest = None

    # Identity fixture
    def fixture(*arg, **kw):
        if arg and callable(arg[0]):
            return arg[0]
        return fixture


# --- Add numpy and ODL to all doctests ---


@fixture(autouse=True)
def _add_doctest_np_odl(doctest_namespace):
    doctest_namespace['np'] = np
    doctest_namespace['odl'] = odl


# --- Ignored tests due to missing modules ---


this_dir = path.dirname(__file__)
odl_root = path.abspath(path.join(this_dir, '..', '..'))
collect_ignore = [path.join(odl_root, 'setup.py'),
                  path.join(odl_root, 'odl', 'contrib')]


# Add example directories to `collect_ignore`
def find_example_dirs():
    dirs = []
    for dirpath, dirnames, _ in os.walk(odl_root):
        if 'examples' in dirnames:
            dirs.append(path.join(dirpath, 'examples'))
    return dirs


collect_ignore.extend(find_example_dirs())


if not PYFFTW_AVAILABLE:
    collect_ignore.append(
        path.join(odl_root, 'odl', 'trafos', 'backends', 'pyfftw_bindings.py')
    )
if not PYWT_AVAILABLE:
    collect_ignore.append(
        path.join(odl_root, 'odl', 'trafos', 'backends', 'pywt_bindings.py')
    )
    # Currently `pywt` is the only implementation
    collect_ignore.append(
        path.join(odl_root, 'odl', 'trafos', 'wavelet.py')
    )


# --- Command-line options --- #


def pytest_addoption(parser):
    suite_help = (
        'enable an opt-in test suite NAME. '
        'Available suites: largescale, examples, doc_doctests'
    )
    parser.addoption(
        '-S',
        '--suite',
        nargs='*',
        metavar='NAME',
        type=str,
        default=[],
        help=suite_help,
    )


def pytest_configure(config):
    # Register an additional marker
    config.addinivalue_line(
        'markers', 'suite(name): mark test to belong to an opt-in suite'
    )


def pytest_runtest_setup(item):
    suites = [mark.args[0] for mark in item.iter_markers(name='suite')]
    if suites:
        if not any(val in suites for val in item.config.getoption('-S')):
            pytest.skip('test not in suites {!r}'.format(suites))


# Remove duplicates
collect_ignore = list(set(collect_ignore))
collect_ignore = [path.normcase(ignored) for ignored in collect_ignore]


# NB: magical `path` param name is needed
def pytest_ignore_collect(path, config):
    normalized = os.path.normcase(str(path))
    return any(normalized.startswith(ignored) for ignored in collect_ignore)


# --- Reusable fixtures --- #

# NOTE: All global fixtures need to be prefixed with `odl_` to make them
# non-conflicting with other packages' fixture names, since these fixtures
# are exposed globally across packages by setuptools.

# Simple ones, use helper
odl_tspace_impl = simple_fixture(name='tspace_impl',
                                 params=tensor_space_impl_names())

real_floating_dtypes = FLOAT_DTYPES
odl_real_floating_dtype = simple_fixture(name='dtype',
                                    params=real_floating_dtypes)

complex_floating_dtypes = COMPLEX_DTYPES
odl_complex_floating_dtype = simple_fixture(name='dtype',
                                    params=complex_floating_dtypes)

floating_dtypes = FLOAT_DTYPES + COMPLEX_DTYPES
odl_floating_dtype = simple_fixture(name='dtype',
                                    params=floating_dtypes)

scalar_dtypes = INTEGER_DTYPES + FLOAT_DTYPES + COMPLEX_DTYPES
odl_scalar_dtype = simple_fixture(name='dtype',
                                  params=scalar_dtypes)


IMPL_DEVICE_PAIRS = []
    
for impl in tensor_space_impl_names():
    array_backend = lookup_array_backend(impl)
    for device in array_backend.available_devices:
        IMPL_DEVICE_PAIRS.append((impl, device))

odl_impl_device_pairs = simple_fixture(name='impl_device', params=IMPL_DEVICE_PAIRS)

if 'pytorch' in tensor_space_impl_names():
    CUDA_DEVICES = []
    for device in lookup_array_backend('pytorch').available_devices:
        CUDA_DEVICES.append(device)

    cuda_device = simple_fixture(name='cuda_device', params=CUDA_DEVICES)

odl_elem_order = simple_fixture(name='order', params=['C'])

odl_reduction = simple_fixture('reduction', ['sum', 'prod', 'min', 'max'])

# More complicated ones with non-trivial documentation
arithmetic_op_par = [operator.add,
                     operator.truediv,
                     operator.mul,
                     operator.sub,
                     operator.iadd,
                     operator.itruediv,
                     operator.imul,
                     operator.isub]
arithmetic_op_ids = [" op = '{}' ".format(op)
                     for op in ['+', '/', '*', '-', '+=', '/=', '*=', '-=']]


@fixture(ids=arithmetic_op_ids, params=arithmetic_op_par)
def odl_arithmetic_op(request):
    """An arithmetic operator, e.g. +, -, // etc."""
    return request.param



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Figures of Merit (FOMs) for measuring image quality without a reference."""

import odl
import numpy as np

__all__ = ('estimate_noise_std',)


def estimate_noise_std(img, average=True):
    """Estimate standard deviation of noise in ``img``.

    The algorithm, given in [Immerkaer1996], estimates the noise in an image.

    Parameters
    ----------
    img : `odl.Tensor`
        Array to estimate noise in.
    average : bool
        If ``True``, return the mean noise in the image, otherwise give a
        pointwise estimate.

    Returns
    -------
    noise : float

    Examples
    --------
    Create image with noise 1.0, verify result

    >>> space = odl.rn((10, 10))
    >>> img = space.element(np.random.randn(*space.shape))
    >>> result = estimate_noise_std(img)  # should be about 1

    Also works with higher dimensional arrays

    >>> space = odl.rn((3, 3, 3))
    >>> img = space.element(np.random.randn(*space.shape))
    >>> result = estimate_noise_std(img)

    The method can also estimate the noise pointwise (but with high
    uncertainty):

    >>> result = estimate_noise_std(img, average=False)

    References
    ----------
    [Immerkaer1996] Immerkaer, J. *Fast Noise Variance Estimation*.
    Computer Vision and Image Understanding, 1996.
    """
    import scipy.signal
    import functools

    space = img.space
    backend = space.array_backend
    device = space.device
    ns = backend.array_namespace

    img = img.astype('float32').asarray()

    M = backend.array_constructor(
            functools.reduce(np.add.outer, [[-1, 2, -1]] * img.ndim)
          , device=device)

    if average:
        # TODO (Justus) it does not really make sense to use FFT for convolving
        # with a small, fixed kernel
        convolved = scipy.signal.fftconvolve(img, M, mode='valid')
        conv_var = ns.sum(convolved ** 2) / convolved.size
    else:
        convolved = scipy.signal.fftconvolve(img, M, mode='same')
        conv_var = convolved ** 2

    scale = ns.sum(ns.square(M))
    sigma = ns.sqrt(conv_var / scale)

    return sigma if average else space.element(sigma)


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Functions for folders and files."""

from __future__ import print_function

from builtins import super

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import scipy.signal
from skimage.io import imsave

import odl

__all__ = ('total_variation', 'TotalVariationNonNegative', 'bregman',
           'save_image', 'save_signal', 'divide_1Darray_equally', 'Blur2D',
           'KullbackLeiblerSmooth')


def save_image(image, name, folder, fignum, cmap='gray', clim=None):
    matplotlib.rc('text', usetex=False)

    fig = plt.figure(fignum)
    plt.clf()

    image.show(name, cmap=cmap, fig=fig)
    fig.savefig('{}/{}_fig.png'.format(folder, name), bbox_inches='tight')

    if clim is None:
        x = image - np.min(image)
        if np.max(x) > 1e-4:
            x /= np.max(x)
    else:
        x = (image - clim[0]) / (clim[1] - clim[0])

    x = np.minimum(np.maximum(x, 0), 1)

    imsave('{}/{}.png'.format(folder, name), np.rot90(x, 1))


def save_signal(signal, name, folder, fignum):
    matplotlib.rc('text', usetex=False)

    fig = plt.figure(fignum)
    plt.clf()

    signal.show(name, fig=fig)
    fig.savefig('{}/{}_fig.png'.format(folder, name), bbox_inches='tight')


def bregman(f, v, subgrad):
    return (odl.functionals.FunctionalQuadraticPerturb(f, linear_term=-subgrad) -
            f(v) + subgrad.inner(v))


def partition_1d(arr, slices):
    return tuple(arr[slc] for slc in slices)


def partition_equally_1d(arr, nparts, order='interlaced'):
    if order == 'block':
        stride = int(np.ceil(arr.size / nparts))
        slices = [slice(i * stride, (i + 1) * stride) for i in range(nparts)]
    elif order == 'interlaced':
        slices = [slice(i, len(arr), nparts) for i in range(nparts)]
    else:
        raise ValueError

    return partition_1d(arr, tuple(slices))


def divide_1Darray_equally(ind, nsub):
    """Divide an array into equal chunks to be used for instance in OSEM.

    Parameters
    ----------
    ind : ndarray
        input array
    nsubsets : int
        number of subsets to be divided into

    Returns
    -------
    sub2ind : list
        list of indices for each subset
    ind2sub : list
        list of subsets for each index
    """

    n_ind = len(ind)
    sub2ind = partition_equally_1d(ind, nsub, order='interlaced')

    ind2sub = []
    for i in range(n_ind):
        ind2sub.append([])

    for i in range(nsub):
        for j in sub2ind[i]:
            ind2sub[j].append(i)

    return (sub2ind, ind2sub)


def total_variation(domain, grad=None):
    """Total variation functional.

    Parameters
    ----------
    domain : odlspace
        domain of TV functional
    grad : gradient operator, optional
        Gradient operator of the total variation functional. This may be any
        linear operator and thereby generalizing TV. default=forward
        differences with Neumann boundary conditions

    Examples
    --------
    Check that the total variation of a constant is zero

    >>> import odl.contrib.solvers.spdhg as spdhg, odl
    >>> space = odl.uniform_discr([0, 0], [3, 3], [3, 3])
    >>> tv = spdhg.total_variation(space)
    >>> x = space.one()
    >>> tv(x) < 1e-10
    True
    """

    if grad is None:
        grad = odl.Gradient(domain, method='forward', pad_mode='symmetric')
        grad.norm = 2 * np.sqrt(sum(1 / grad.domain.cell_sides**2))
    else:
        grad = grad

    f = odl.functionals.GroupL1Norm(grad.range, exponent=2)

    return f * grad


class TotalVariationNonNegative(odl.functionals.Functional):
    """Total variation function with nonnegativity constraint and strongly
    convex relaxation.

    In formulas, this functional may represent

        alpha * |grad x|_1 + char_fun(x) + beta/2 |x|^2_2

    with regularization parameter alpha and strong convexity beta. In addition,
    the nonnegativity constraint is achieved with the characteristic function

        char_fun(x) = 0 if x >= 0 and infty else.

    Parameters
    ----------
    domain : odlspace
        domain of TV functional
    alpha : scalar, optional
        Regularization parameter, positive
    prox_options : dict, optional
        name: string, optional
            name of the method to perform the prox operator, default=FGP
        warmstart: boolean, optional
            Do you want a warm start, i.e. start with the dual variable
            from the last call? default=True
        niter: int, optional
            number of iterations per call, default=5
        p: array, optional
            initial dual variable, default=zeros
    grad : gradient operator, optional
        Gradient operator to be used within the total variation functional.
        default=see TV
    """

    def __init__(self, domain, alpha=1, prox_options={}, grad=None,
                 strong_convexity=0):
        """
        """

        self.strong_convexity = strong_convexity

        if 'name' not in prox_options:
            prox_options['name'] = 'FGP'
        if 'warmstart' not in prox_options:
            prox_options['warmstart'] = True
        if 'niter' not in prox_options:
            prox_options['niter'] = 5
        if 'p' not in prox_options:
            prox_options['p'] = None
        if 'tol' not in prox_options:
            prox_options['tol'] = None

        self.prox_options = prox_options

        self.alpha = alpha
        self.tv = total_variation(domain, grad=grad)
        self.grad = self.tv.right
        self.nn = odl.functionals.IndicatorBox(domain, 0, np.inf)
        self.l2 = 0.5 * odl.functionals.L2NormSquared(domain)
        self.proj_P = self.tv.left.convex_conj.proximal(0)
        self.proj_C = self.nn.proximal(1)

        super().__init__(space=domain, linear=False, grad_lipschitz=0)

    def __call__(self, x):
        """Evaluate functional.

        Examples
        --------
        Check that the total variation of a constant is zero

        >>> import odl.contrib.solvers.spdhg as spdhg, odl
        >>> space = odl.uniform_discr([0, 0], [3, 3], [3, 3])
        >>> tvnn = spdhg.TotalVariationNonNegative(space, alpha=2)
        >>> x = space.one()
        >>> tvnn(x) < 1e-10
        True

        Check that negative functions are mapped to infty

        >>> import odl.contrib.solvers.spdhg as spdhg, odl, numpy as np
        >>> space = odl.uniform_discr([0, 0], [3, 3], [3, 3])
        >>> tvnn = spdhg.TotalVariationNonNegative(space, alpha=2)
        >>> x = -space.one()
        >>> np.isinf(tvnn(x))
        True
        """

        nn = self.nn(x)

        if nn is np.inf:
            return nn
        else:
            out = self.alpha * self.tv(x) + nn
            if self.strong_convexity > 0:
                out += self.strong_convexity * self.l2(x)
            return out

    def proximal(self, sigma):
        """Prox operator of TV. It allows the proximal step length to be a
        vector of positive elements.

        Examples
        --------
        Check that the proximal operator is the identity for sigma=0

        >>> import odl.contrib.solvers.spdhg as spdhg, odl, numpy as np
        >>> space = odl.uniform_discr([0, 0], [3, 3], [3, 3])
        >>> tvnn = spdhg.TotalVariationNonNegative(space, alpha=2)
        >>> x = -space.one()
        >>> y = tvnn.proximal(0)(x)
        >>> (y-x).norm() < 1e-10
        True

        Check that negative functions are mapped to 0

        >>> import odl.contrib.solvers.spdhg as spdhg, odl, numpy as np
        >>> space = odl.uniform_discr([0, 0], [3, 3], [3, 3])
        >>> tvnn = spdhg.TotalVariationNonNegative(space, alpha=2)
        >>> x = -space.one()
        >>> y = tvnn.proximal(0.1)(x)
        >>> y.norm() < 1e-10
        True
        """

        if sigma == 0:
            return odl.IdentityOperator(self.domain)

        else:
            def tv_prox(z, out=None):

                if out is None:
                    out = z.space.zero()

                opts = self.prox_options

                sigma_ = np.copy(sigma)
                z_ = z.copy()

                if self.strong_convexity > 0:
                    sigma_ /= (1 + sigma * self.strong_convexity)
                    z_ /= (1 + sigma * self.strong_convexity)

                if opts['name'] == 'FGP':
                    if opts['warmstart']:
                        if opts['p'] is None:
                            opts['p'] = self.grad.range.zero()

                        p = opts['p']
                    else:
                        p = self.grad.range.zero()

                    sigma_sqrt = np.sqrt(sigma_)

                    z_ /= sigma_sqrt
                    grad = sigma_sqrt * self.grad
                    grad.norm = sigma_sqrt * self.grad.norm
                    niter = opts['niter']
                    alpha = self.alpha
                    out[:] = fgp_dual(p, z_, alpha, niter, grad, self.proj_C,
                                      self.proj_P, tol=opts['tol'])

                    out *= sigma_sqrt

                    return out

                else:
                    raise NotImplementedError('Not yet implemented')

            return tv_prox


def fgp_dual(p, data, alpha, niter, grad, proj_C, proj_P, tol=None, **kwargs):
    """Computes a solution to the ROF problem with the fast gradient
    projection algorithm.

    Parameters
    ----------
    p : np.array
        dual initial variable
    data : np.array
        noisy data / proximal point
    alpha : float
        regularization parameter
    niter : int
        number of iterations
    grad : instance of gradient class
        class that supports grad(x), grad.adjoint(x), grad.norm
    proj_C : function
        projection onto the constraint set of the primal variable,
        e.g. non-negativity
    proj_P : function
        projection onto the constraint set of the dual variable,
        e.g. norm <= 1
    tol : float (optional)
        nonnegative parameter that gives the tolerance for convergence. If set
        None, then the algorithm will run for a fixed number of iterations

    Other Parameters
    ----------------
    callback : callable, optional
        Function called with the current iterate after each iteration.
    """

    # Callback object
    callback = kwargs.pop('callback', None)
    if callback is not None and not callable(callback):
        raise TypeError('`callback` {} is not callable'.format(callback))

    factr = 1 / (grad.norm**2 * alpha)

    q = p.copy()
    x = data.space.zero()

    t = 1.

    if tol is None:
        def convergence_eval(p1, p2):
            return False
    else:
        def convergence_eval(p1, p2):
            return (p1 - p2).norm() / p1.norm() < tol

    pnew = p.copy()

    if callback is not None:
        callback(p)

    for k in range(niter):
        t0 = t
        grad.adjoint(q, out=x)
        proj_C(data - alpha * x, out=x)
        grad(x, out=pnew)
        pnew *= factr
        pnew += q

        proj_P(pnew, out=pnew)

        converged = convergence_eval(p, pnew)

        if not converged:
            # update step size
            t = (1 + np.sqrt(1 + 4 * t0 ** 2)) / 2.

            # calculate next iterate
            q[:] = pnew + (t0 - 1) / t * (pnew - p)

        p[:] = pnew

        if converged:
            t = None
            break

        if callback is not None:
            callback(p)

    # get current image estimate
    x = proj_C(data - alpha * grad.adjoint(p))

    return x


class Blur2D(odl.Operator):
    """Blur operator"""

    def __init__(self, domain, kernel, boundary_condition='wrap'):
        """Initialize a new instance.
        """

        super().__init__(domain=domain, range=domain, linear=True)

        self.__kernel = kernel
        self.__boundary_condition = boundary_condition

    @property
    def kernel(self):
        return self.__kernel

    @property
    def boundary_condition(self):
        return self.__boundary_condition

    def _call(self, x, out):
        out[:] = scipy.signal.convolve2d(x, self.kernel, mode='same',
                                         boundary='wrap')

    @property
    def gradient(self):
        raise NotImplementedError('No yet implemented')

    @property
    def adjoint(self):
        adjoint_kernel = self.kernel.copy().conj()
        adjoint_kernel = np.fliplr(np.flipud(adjoint_kernel))

        return Blur2D(self.domain, adjoint_kernel, self.boundary_condition)

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}({!r}, {!r}, {!r})'.format(
            self.__class__.__name__, self.domain, self.kernel,
            self.boundary_condition)


class KullbackLeiblerSmooth(odl.functionals.Functional):

    r"""The smooth Kullback-Leibler divergence functional.

    Notes
    -----
    If the functional is defined on an :math:`\mathbb{R}^n`-like space, the
    smooth Kullback-Leibler functional :math:`\phi` is defined as

    .. math::
        \phi(x) = \sum_{i=1}^n \begin{cases}
                x + r - y + y * \log(y / (x + r))
                    & \text{if $x \geq 0$} \
                (y / (2 * r^2)) * x^2 + (1 - y / r) * x + r - b +
                    b * \log(b / r) & \text{else}
                                 \end{cases}

    where all variables on the right hand side of the equation have a subscript
    i which is omitted for readability.

    References
    ----------
    [CERS2017] A. Chambolle, M. J. Ehrhardt, P. Richtarik and C.-B. Schoenlieb,
    *Stochastic Primal-Dual Hybrid Gradient Algorithm with Arbitrary Sampling
    and Imaging Applications*. ArXiv: http://arxiv.org/abs/1706.04957 (2017).
    """

    def __init__(self, space, data, background):
        """Initialize a new instance.

        Parameters
        ----------
        space : `DiscretizedSpace` or `TensorSpace`
            Domain of the functional.
        data : ``space`` `element-like`
            Data vector which has to be non-negative.
        background : ``space`` `element-like`
            Background vector which has to be non-negative.
        """

        self.strong_convexity = 0

        if background.ufuncs.less_equal(0).ufuncs.sum() > 0:
            raise NotImplementedError('Background must be positive')

        super().__init__(space=space, linear=False,
                         grad_lipschitz=np.max(data / background ** 2))

        if data not in self.domain:
            raise ValueError('`data` not in `domain`'
                             ''.format(data, self.domain))

        self.__data = data
        self.__background = background

    @property
    def data(self):
        """The data in the Kullback-Leibler functional."""
        return self.__data

    @property
    def background(self):
        """The background in the Kullback-Leibler functional."""
        return self.__background

    def _call(self, x):
        """Return the KL-diveregnce in the point ``x``.

        If any components of ``x`` is non-positive, the value is positive
        infinity.
        """
        y = self.data
        r = self.background

        obj = self.domain.zero()

        # x + r - y + y * log(y / (x + r)) = x - y * log(x + r) + c1
        # with c1 = r - y + y * log y
        i = x.ufuncs.greater_equal(0)
        obj[i] = x[i] + r[i] - y[i]

        j = y.ufuncs.greater(0)
        k = i.ufuncs.logical_and(j)
        obj[k] += y[k] * (y[k] / (x[k] + r[k])).ufuncs.log()

        # (y / (2 * r^2)) * x^2 + (1 - y / r) * x + r - b + b * log(b / r)
        # = (y / (2 * r^2)) * x^2 + (1 - y / r) * x + c2
        # with c2 = r - b + b * log(b / r)
        i = i.ufuncs.logical_not()
        obj[i] += (y[i] / (2 * r[i]**2) * x[i]**2 + (1 - y[i] / r[i]) * x[i] +
                   r[i] - y[i])

        k = i.ufuncs.logical_and(j)
        obj[k] += y[k] * (y[k] / r[k]).ufuncs.log()

        return obj.inner(self.domain.one())

    @property
    def gradient(self):
        """Gradient operator of the functional.
        """
        raise NotImplementedError('No yet implemented')

    @property
    def proximal(self):
        """Return the `proximal factory` of the functional.
        """
        raise NotImplementedError('No yet implemented')

    @property
    def convex_conj(self):
        """The convex conjugate functional of the KL-functional."""
        return KullbackLeiblerSmoothConvexConj(self.domain, self.data,
                                               self.background)

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}({!r}, {!r}, {!r})'.format(
            self.__class__.__name__, self.domain, self.data, self.background)


class KullbackLeiblerSmoothConvexConj(odl.functionals.Functional):

    r"""The convex conj of the smooth Kullback-Leibler divergence functional.

    Notes
    -----
    If the functional is defined on an :math:`\mathbb{R}^n`-like space, the
    convex conjugate of the smooth Kullback-Leibler functional :math:`\phi^*`
    is defined as

    .. math::
        \phi^*(x) = \sum_{i=1}^n \begin{cases}
                r^2 / (2 * y) * x^2 + (r - r^2 / y) * x + r^2 / (2 * y) +
                    3 / 2 * y - 2 * r - y * log(y / r)
                    & \text{if $x < 1 - y / r$} \
                - r * x - y * log(1 - x)
                    & \text{if $1 - y / r <= x < 1} \
                + \infty
                    & \text{else}
                                 \end{cases}

    where all variables on the right hand side of the equation have a subscript
    :math:`i` which is omitted for readability.

    References
    ----------
    [CERS2017] A. Chambolle, M. J. Ehrhardt, P. Richtarik and C.-B. Schoenlieb,
    *Stochastic Primal-Dual Hybrid Gradient Algorithm with Arbitrary Sampling
    and Imaging Applications*. ArXiv: http://arxiv.org/abs/1706.04957 (2017).
    """

    def __init__(self, space, data, background):
        """Initialize a new instance.

        Parameters
        ----------
        space : `DiscretizedSpace` or `TensorSpace`
            Domain of the functional.
        data : ``space`` `element-like`
            Data vector which has to be non-negative.
        background : ``space`` `element-like`
            Background vector which has to be non-negative.
        """

        if background.ufuncs.less_equal(0).ufuncs.sum() > 0:
            raise NotImplementedError('Background must be positive')

        super().__init__(space=space, linear=False,
                         grad_lipschitz=np.inf)

        if data is not None and data not in self.domain:
            raise ValueError('`data` not in `domain`'
                             ''.format(data, self.domain))

        self.__data = data
        self.__background = background

        if np.min(self.data) == 0:
            self.strong_convexity = np.inf
        else:
            self.strong_convexity = np.min(self.background**2 / self.data)

    @property
    def data(self):
        """The data in the Kullback-Leibler functional."""
        return self.__data

    @property
    def background(self):
        """The background in the Kullback-Leibler functional."""
        return self.__background

    def _call(self, x):
        """Return the value in the point ``x``.

        If any components of ``x`` is larger than or equal to 1, the value is
        positive infinity.
        """

        # TODO: cover properly the case data = 0

        y = self.data
        r = self.background

        # if any element is greater or equal to one
        if x.ufuncs.greater_equal(1).ufuncs.sum() > 0:
            return np.inf

        obj = self.domain.zero()

        # out = sum(f)
        # f =
        #     if x < 1 - y / r:
        #         r^2 / (2 * y) * x^2 + (r - r^2 / y) * x + r^2 / (2 * y) +
        #             3 / 2 * y - 2 * r - y * log(y / r)
        #     if x >= 1 - y / r:
        #         - r * x - y * log(1 - x)
        i = x.ufuncs.less(1 - y / r)
        ry = r[i]**2 / y[i]
        obj[i] += (ry / 2 * x[i]**2 + (r[i] - ry) * x[i] + ry / 2 +
                   3 / 2 * y[i] - 2 * r[i])

        j = y.ufuncs.greater(0)
        k = i.ufuncs.logical_and(j)
        obj[k] -= y[k] * (y[k] / r[k]).ufuncs.log()

        i = i.ufuncs.logical_not()
        obj[i] -= r[i] * x[i]

        k = i.ufuncs.logical_and(j)
        obj[k] -= y[k] * (1 - x[k]).ufuncs.log()

        return obj.inner(self.domain.one())

    @property
    def gradient(self):
        """Gradient operator of the functional."""
        raise NotImplementedError('No yet implemented')

    @property
    def proximal(self):

        space = self.domain
        y = self.data
        r = self.background

        class ProxKullbackLeiblerSmoothConvexConj(odl.Operator):
            """Proximal operator of the convex conjugate of the smooth
            Kullback-Leibler functional.
            """

            def __init__(self, sigma):
                """Initialize a new instance.

                Parameters
                ----------
                sigma : positive float
                    Step size parameter
                """
                self.sigma = float(sigma)
                self.background = r
                self.data = y
                super().__init__(domain=space, range=space, linear=False)

            def _call(self, x, out):

                s = self.sigma
                y = self.data
                r = self.background

                sr = s * r
                sy = s * y

                # out =
                #    if x  < 1 - y / r:
                #        (y * x - s * r * y + s * r**2) / (y + s * r**2)
                #    if x >= 1 - y / r:
                #        0.5 * (x + s * r + 1 -
                #            sqrt((x + s * r - 1)**2 + 4 * s * y)

                i = x.ufuncs.less(1 - y / r)
                # TODO: This may be faster without indexing on the GPU?
                out[i] = ((y[i] * x[i] - sr[i] * y[i] + sr[i] * r[i]) /
                          (y[i] + sr[i] * r[i]))

                i.ufuncs.logical_not(out=i)
                out[i] = (x[i] + sr[i] + 1 -
                          ((x[i] + sr[i] - 1) ** 2 + 4 * sy[i]).ufuncs.sqrt())
                out[i] /= 2

                return out

        return ProxKullbackLeiblerSmoothConvexConj

    @property
    def convex_conj(self):
        """The convex conjugate functional of the smooth KL-functional."""
        return KullbackLeiblerSmooth(self.domain, self.data,
                                     self.background)

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}({!r}, {!r}, {!r})'.format(
            self.__class__.__name__, self.domain, self.data, self.background)
